"""Tests file operations."""
# pylint: disable=redefined-outer-name

import json
import sys
import tempfile
import uuid as uuidlib
from datetime import datetime
from pathlib import Path
from pathlib import PurePosixPath

import pytest
import responses
from loguru import logger as log
from spectrumx import Client
from spectrumx.api.sds_files import delete_file
from spectrumx.gateway import API_TARGET_VERSION
from spectrumx.ops.files import get_file_permissions

from tests.conftest import get_content_check_endpoint
from tests.conftest import get_files_endpoint

log.trace("Placeholder log to avoid reimporting or resolving unused import warnings.")


def _download_file_endpoint(client: Client, file_id: str) -> str:
    """Returns the endpoint for downloading a file."""
    return (
        client.base_url + f"/api/{API_TARGET_VERSION}/assets/files/{file_id}/download/"
    )


def _get_file_id_endpoint(client: Client, file_id: str) -> str:
    """Returns the endpoint for getting a file."""
    return client.base_url + f"/api/{API_TARGET_VERSION}/assets/files/{file_id}/"


@pytest.mark.linux
@pytest.mark.darwin
def test_get_file_permissions(temp_file_empty: Path) -> None:
    """Test get_file_permissions for many permission combinations."""
    chmod_combos = {
        "---------": 0o000,
        "--------x": 0o001,
        "r-xr-xr--": 0o554,
        "r--rw-r--": 0o464,
        "rw-rw-r--": 0o664,
        "rwx------": 0o700,
        "rwxrwxrwx": 0o777,
    }
    for perm_string, chmod in chmod_combos.items():
        temp_file_empty.chmod(chmod)
        assert get_file_permissions(temp_file_empty) == perm_string


@pytest.mark.win32
def test_get_file_permissions_win32(temp_file_empty: Path) -> None:
    """
    Test get_file_permissions for many permission combinations.

    Windows only allows setting read or write permissions on a file.
    """
    chmod_combos = {
        "r--r--r--": 0o400,
        "r--r--r--": 0o440,  # noqa: F601
        "r--r--r--": 0o444,  # noqa: F601
        "rw-rw-rw-": 0o600,
        "rw-rw-rw-": 0o660,  # noqa: F601
        "rw-rw-rw-": 0o666,  # noqa: F601
    }
    for perm_string, chmod in chmod_combos.items():
        temp_file_empty.chmod(chmod)
        assert get_file_permissions(temp_file_empty) == perm_string


def test_get_file_by_id(client: Client, responses: responses.RequestsMock) -> None:
    """Given a file ID, the client must return the file."""
    uuid = uuidlib.uuid4()
    url: str = get_files_endpoint(client) + f"{uuid.hex}/"
    responses.get(
        url=url,
        status=200,
        json={
            "created_at": "2021-10-01T12:00:00Z",
            "directory": "/my/files/are/here/",
            "expiration_date": "2021-10-01T12:12:00Z",
            "media_type": "text/plain",
            "name": "file.txt",
            "permissions": "rw-rw-r--",
            "size": 321,
            "updated_at": "2021-10-01T12:00:00Z",
            "uuid": uuid.hex,
        },
    )
    client.dry_run = False  # to test the actual request
    file_sample = client.get_file(file_uuid=uuid.hex)
    assert file_sample.uuid == uuid


def test_file_get_returns_valid(
    client: Client,
) -> None:
    """The get_file method must return a valid File instance.

    Note the file may not exist locally, but the File instance can still be valid.
        That is the case for dry-runs - which use sample files - and for files fetched
        from the server that have not yet been downloaded.
    """

    file_id = uuidlib.uuid4()
    file_size = 888
    client.dry_run = True
    assert client.dry_run is True, "Dry run must be enabled for this test."
    file_sample = client.get_file(file_uuid=file_id.hex)
    assert file_sample.is_sample is True, "The file must be a sample file"  # pyright: ignore[reportPrivateUsage]
    assert file_sample.uuid is not None
    assert file_sample.name.startswith("dry-run-")
    assert file_sample.media_type == "text/plain"
    assert file_sample.size == file_size
    assert file_sample.directory == PurePosixPath("sds-files/dry-run/")
    assert file_sample.permissions == "rw-rw-r--"
    assert isinstance(file_sample.created_at, datetime)
    assert isinstance(file_sample.updated_at, datetime)
    assert isinstance(file_sample.expiration_date, datetime), (
        "Expiration date should be a datetime"
    )
    assert file_sample.expiration_date > file_sample.created_at


def test_file_upload_returns_file(
    client: Client, temp_file_with_text_contents: Path, tmp_path: Path
) -> None:
    """The upload_file method must return a valid File instance."""
    test_file_size = temp_file_with_text_contents.stat().st_size
    client.dry_run = True
    assert client.dry_run is True, "Dry run must be enabled for this test."
    file_sample = client.upload_file(
        local_file=temp_file_with_text_contents,
        sds_path=PurePosixPath("/my/upload/location"),
    )
    assert file_sample.is_sample is False, (
        "The file must be a real file on disk (not a sample), even for this test"
    )  # pyright: ignore[reportPrivateUsage]
    assert temp_file_with_text_contents == file_sample.local_path, (
        "Local path does not match"
    )
    assert file_sample.uuid is None, "A local file must not have a UUID"
    assert file_sample.name is not None, "Expected a file name"
    assert file_sample.media_type == "text/plain", "Expected media type 'text/plain'"
    assert file_sample.size == test_file_size, "Expected the test file to be 4030 bytes"
    assert str(tmp_path) in str(file_sample.local_path), (
        "Expected the temp file directory"
    )
    assert file_sample.directory == PurePosixPath("/my/upload/location")
    expected_permissions = "rw-rw-rw-" if sys.platform == "win32" else "rw-r--r--"
    assert file_sample.permissions == expected_permissions

    assert isinstance(file_sample.created_at, datetime)
    assert isinstance(file_sample.updated_at, datetime)
    assert file_sample.expiration_date is None, (
        "Local files should not have an expiration date"
    )


def test_large_file_upload_mocked(
    client: Client, responses: responses.RequestsMock, temp_large_binary_file: Path
) -> None:
    """Test the upload_file method with mocked responses."""
    file_id = uuidlib.uuid4()
    file_size = temp_large_binary_file.stat().st_size
    client.dry_run = False  # calls are mocked, but we want to test the actual requests
    mocked_file_content_check_json = {
        "file_contents_exist_for_user": False,
        "file_exists_in_tree": False,
        "user_mutable_attributes_differ": True,
    }
    responses.add(
        method=responses.POST,
        url=get_content_check_endpoint(client),
        status=201,
        json=mocked_file_content_check_json,
    )
    mocked_upload_json = {
        "uuid": file_id.hex,
        "name": temp_large_binary_file.name,
        "media_type": "text/plain",
        "size": file_size,
        "directory": "/my/custom/sds/location",
        "permissions": "rw-r--r--",
        "created_at": "2024-12-01T12:00:00Z",
        "updated_at": "2024-12-01T12:00:00Z",
        "expiration_date": "2026-12-01T12:00:00Z",
    }
    responses.add(
        method=responses.POST,
        url=get_files_endpoint(client),
        status=201,
        json=mocked_upload_json,
    )
    # run the test
    file_sample = client.upload_file(
        local_file=temp_large_binary_file,
        sds_path=PurePosixPath(mocked_upload_json["directory"]),
    )
    assert file_sample.uuid == file_id, "UUID not as mocked."
    assert file_sample.name == mocked_upload_json["name"]
    assert file_sample.media_type == mocked_upload_json["media_type"]
    assert file_sample.size == file_size
    assert file_sample.directory == PurePosixPath(mocked_upload_json["directory"])
    assert file_sample.permissions == mocked_upload_json["permissions"]
    assert isinstance(file_sample.created_at, datetime)
    assert isinstance(file_sample.updated_at, datetime)


def test_download_file_contents(
    client: Client, responses: responses.RequestsMock
) -> None:
    """The download_file_contents method must create a file with the file contents."""
    client.dry_run = False  # calls are mocked, but we want to test the actual requests

    # file properties
    file_id = uuidlib.uuid4()
    long_file_contents = (
        "File start\n" + "Sample file contents - " * 1000 + "\nFile end"
    )
    num_bytes = len(long_file_contents.encode("utf-8"))

    file_metadata = {
        "uuid": file_id.hex,
        "name": "downloaded-file.txt",
        "media_type": "text/plain",
        "size": num_bytes,
        "directory": "/my/download/location",
        "permissions": "rw-r--r--",
        "created_at": "2024-12-01T12:00:00Z",
        "updated_at": "2024-12-01T12:00:00Z",
        "expiration_date": "2026-12-01T12:00:00Z",
    }

    # mock the file metadata endpoint
    responses.add(
        method=responses.GET,
        url=_get_file_id_endpoint(client, file_id=file_id.hex),
        status=200,
        body=json.dumps(file_metadata),
    )

    # mock the file download endpoint
    responses.add(
        method=responses.GET,
        url=_download_file_endpoint(client, file_id=file_id.hex),
        status=200,
        body=long_file_contents,
    )

    # run the test
    downloaded_file = client.download_file(file_uuid=file_id.hex)
    downloaded_path = downloaded_file.local_path
    assert downloaded_path is not None, "Returned path must not be None"
    assert downloaded_path.exists(), "File was not downloaded to returned path"
    assert downloaded_path.is_file(), "Returned path must be a file"
    assert downloaded_path.stat().st_size == num_bytes, (
        f"File size must match. Got {num_bytes} B, "
        f"expected {downloaded_file.stat().st_size} B"
    )

    # cleanup
    downloaded_path.unlink(missing_ok=False)


def test_download_file_to_path(
    client: Client, responses: responses.RequestsMock
) -> None:
    """The download method must respect the given path, creating any parent dirs."""
    client.dry_run = False  # calls are mocked, but we want to test the actual requests

    # file properties
    file_id = uuidlib.uuid4()
    file_contents = "File start\n" + "Sample file contents" + "\nFile end"
    num_bytes = len(file_contents.encode("utf-8"))
    random_str = uuidlib.uuid4().hex
    parent_dir = Path(tempfile.gettempdir()) / random_str
    expected_path = parent_dir / "downloaded-file.txt"

    # mock the file metadata endpoint
    file_metadata = {
        "uuid": file_id.hex,
        "name": expected_path.name,
        "media_type": "text/plain",
        "size": num_bytes,
        "directory": "/my/download/location",
        "permissions": "rw-r--r--",
        "created_at": "2024-12-01T12:00:00Z",
        "updated_at": "2024-12-01T12:00:00Z",
        "expiration_date": "2026-12-01T12:00:00Z",
    }
    responses.add(
        method=responses.GET,
        url=_get_file_id_endpoint(client, file_id=file_id.hex),
        status=200,
        body=json.dumps(file_metadata),
    )

    # mock the file download endpoint
    responses.add(
        method=responses.GET,
        url=_download_file_endpoint(client, file_id=file_id.hex),
        status=200,
        body=file_contents,
    )

    # run the test
    assert parent_dir.exists() is False, "Parent dir must not exist before test"
    downloaded_file = client.download_file(
        file_uuid=file_id.hex, to_local_path=expected_path
    )
    downloaded_path = downloaded_file.local_path
    assert downloaded_path == expected_path, "Returned path must match the given path"
    assert expected_path.exists(), "File was not downloaded to given path"
    assert expected_path.is_file(), "Given path must be a file"
    assert expected_path.stat().st_size == num_bytes, (
        f"File size must match. Got {num_bytes} B, "
        f"expected {expected_path.stat().st_size} B"
    )

    # cleanup
    expected_path.unlink(missing_ok=False)
    parent_dir.rmdir()


@responses.activate
def test_delete_file_success(client: Client) -> None:
    """Test successful file deletion."""
    # ARRANGE
    test_uuid = uuidlib.uuid4()
    test_uuid_hex = test_uuid.hex
    client.dry_run = False
    assert client.dry_run is False, "Dry run must be enabled for this test."
    responses.add(
        responses.DELETE,
        f"{get_files_endpoint(client)}{test_uuid_hex}/",
        status=204,
    )

    # ACT
    result = delete_file(client=client, file_uuid=test_uuid)

    # ASSERT
    assert result is True, "Expected deletion to succeed."
    assert len(responses.calls) == 1
    assert responses.calls[0].request.method == "DELETE"
    assert (
        responses.calls[0].request.url
        == f"{get_files_endpoint(client)}{test_uuid_hex}/"
    )


@responses.activate
def test_delete_file_str_uuid(client: Client) -> None:
    """Test file deletion with string UUID."""
    # ARRANGE
    test_uuid = uuidlib.uuid4()
    test_uuid_hex = test_uuid.hex
    client.dry_run = False  # calls are mocked, but we want to test the actual requests
    assert client.dry_run is False, "Dry run must be enabled for this test."
    responses.add(
        responses.DELETE,
        f"{get_files_endpoint(client)}{test_uuid_hex}/",
        status=204,
    )

    # ACT
    result = delete_file(client=client, file_uuid=test_uuid_hex)

    # ASSERT
    assert result is True, "Expected deletion to succeed."
    assert len(responses.calls) == 1, "Expected one DELETE request to be made."
    assert responses.calls[0].request.method == "DELETE"
    assert (
        responses.calls[0].request.url
        == f"{get_files_endpoint(client)}{test_uuid_hex}/"
    )


@responses.activate
def test_delete_file_dry_run(client: Client) -> None:
    """Test file deletion in dry run mode."""
    # ARRANGE
    test_uuid = uuidlib.uuid4()
    assert client.dry_run is True, "Dry run must be enabled for this test."

    # ACT
    result = delete_file(client=client, file_uuid=test_uuid)

    # ASSERT
    assert result is True  # deletion always succeeds in dry run mode
    assert len(responses.calls) == 0



"""Tests for SDS Captures."""

import json
from collections.abc import Generator
from pathlib import Path
from pathlib import PurePosixPath
from typing import Any

import pytest
from loguru import logger as log
from pydantic import BaseModel
from spectrumx.client import Client
from spectrumx.errors import CaptureError
from spectrumx.models.captures import Capture
from spectrumx.models.captures import CaptureType
from spectrumx.utils import get_random_line

from tests.integration.conftest import PassthruEndpoints
from tests.integration.conftest import dir_integration_data

log.trace("Placeholder log avoid reimporting or resolving unused import warnings.")

# paths with test data
drf_channel = "cap-2024-06-27T14-00-00"


@pytest.fixture
def _capture_test(drf_sample_top_level_dir: Path) -> Generator[None]:
    """Choose to run capture test or not, based on existing data."""
    if not drf_sample_top_level_dir.exists():
        pytest.skip(
            "Capture test SKIPPED. Test data does not "
            f"exist at: '{drf_sample_top_level_dir}'"
        )

    # yield control to the test
    yield

    # teardown code for integration tests

    return


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
        ]
    ],
    indirect=True,
)
def test_capture_creation_drf(
    integration_client: Client, drf_sample_top_level_dir: Path
) -> None:
    """Tests creating a Digital-RF capture."""

    # ARRANGE
    cap_data = _upload_drf_capture_test_assets(
        integration_client=integration_client,
        drf_sample_top_level_dir=drf_sample_top_level_dir,
    )

    # ACT

    # create a capture
    capture = integration_client.captures.create(
        top_level_dir=cap_data.capture_top_level,
        channel=cap_data.drf_channel,
        capture_type=CaptureType.DigitalRF,
    )

    # ASSERT

    # basic capture information
    assert capture.uuid is not None, "Capture UUID should not be None"
    assert capture.capture_type == CaptureType.DigitalRF
    assert capture.channel == drf_channel
    assert capture.top_level_dir == cap_data.capture_top_level

    # test capture properties
    assert capture.capture_props["start_bound"] == cap_data.cap_start_bound
    assert capture.capture_props["is_continuous"] == cap_data.cap_is_continuous
    assert (
        capture.capture_props["custom_attrs"]["receiver/info/mboard_serial"]
        == cap_data.cap_serial
    )


class DRFCaptureAssets(BaseModel):
    """Holds Digital-RF capture assets to ease uploads and capture creations."""

    cap_is_continuous: bool
    cap_serial: str
    cap_start_bound: int
    drf_channel: str
    path_after_capture_data: PurePosixPath

    @property
    def capture_top_level(self) -> PurePosixPath:
        """Constructs the full path for the capture on the SDS."""
        return PurePosixPath("/") / self.path_after_capture_data


def _upload_drf_capture_test_assets(
    integration_client: Client,
    drf_sample_top_level_dir: Path,
) -> DRFCaptureAssets:
    """Helper to package and upload Digital-RF capture assets to SDS."""
    cap_start_bound: int = 1_719_499_740
    cap_is_continuous: bool = True
    cap_serial: str = "31649FE"

    # define paths in the context of a capture
    dir_top_level = drf_sample_top_level_dir
    dir_channel = drf_sample_top_level_dir / drf_channel
    dir_metadata = dir_channel / "metadata"
    assert dir_metadata.is_dir(), (
        "Metadata directory should exist; check that "
        f"you have the right paths set: '{dir_metadata}'"
    )

    # suffix path_after_capture_data with a random name to avoid conflicts between runs
    path_after_capture_data = dir_top_level.relative_to(dir_integration_data)
    random_suffix = get_random_line(10, include_punctuation=False)
    path_after_capture_data = (
        PurePosixPath(path_after_capture_data) / f"test-{random_suffix}"
    )

    _upload_assets(
        integration_client=integration_client,
        sds_path=path_after_capture_data,
        local_path=dir_top_level,
    )

    return DRFCaptureAssets(
        cap_is_continuous=cap_is_continuous,
        cap_serial=cap_serial,
        cap_start_bound=cap_start_bound,
        drf_channel=drf_channel,
        path_after_capture_data=path_after_capture_data,
    )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
        ]
    ],
    indirect=True,
)
def test_capture_creation_rh(
    integration_client: Client,
    rh_sample_top_level_dir: Path,
) -> None:
    """Tests creating a RadioHound capture."""

    # ARRANGE
    radiohound_data = load_rh_data(rh_sample_top_level_dir)
    scan_group = radiohound_data["scan_group"]

    # suffix path_after_capture_data with a random name to avoid conflicts between runs
    rel_path_capture = rh_sample_top_level_dir.relative_to(dir_integration_data)
    random_suffix = get_random_line(10, include_punctuation=False)
    rel_path_capture = PurePosixPath(rel_path_capture) / f"test-{random_suffix}"

    _upload_assets(
        integration_client=integration_client,
        sds_path=rel_path_capture,
        local_path=rh_sample_top_level_dir,
    )

    # ACT

    # delete all captures with that scan group to avoid conflicts
    _delete_rh_captures_by_scan_group(
        integration_client=integration_client,
        scan_group=scan_group,
    )

    capture_top_level = PurePosixPath("/") / rel_path_capture
    capture = integration_client.captures.create(
        top_level_dir=capture_top_level,
        scan_group=scan_group,
        capture_type=CaptureType.RadioHound,
    )

    # ASSERT

    # basic capture information
    assert capture.uuid is not None, "Capture UUID should not be None"
    assert capture.capture_type == CaptureType.RadioHound
    assert capture.top_level_dir == capture_top_level

    # test capture metadata
    assert capture.scan_group is not None, "Scan group should not be None"
    assert str(capture.scan_group) == radiohound_data["scan_group"], (
        "Scan group was not set correctly for RadioHound capture"
    )
    assert capture.capture_props, "Capture properties should not be empty"
    assert capture.capture_props == radiohound_data, (
        "Capture props doesn't match the reference data: \n"
        f"'{capture.capture_props}'\n!=\n'{radiohound_data}'"
    )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.capture_listing(),
        ]
    ],
    indirect=True,
)
def test_capture_listing_drf(integration_client: Client) -> None:
    """Tests reading and listing Digital-RF captures."""
    captures = integration_client.captures.listing(
        capture_type=CaptureType.DigitalRF,
    )
    assert len(captures) > 0, "At least one capture should be present"
    for capture in captures:
        assert capture.uuid is not None, "Capture UUID should not be None"
        assert capture.capture_type == CaptureType.DigitalRF
        assert capture.channel is not None, "DigitalRF capture should have a channel"
        assert capture.top_level_dir is not None, "Top level dir should not be None"


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.capture_listing(),
        ]
    ],
    indirect=True,
)
def test_capture_listing_rh(integration_client: Client) -> None:
    """Tests reading and listing RadioHound captures."""
    captures = integration_client.captures.listing(
        capture_type=CaptureType.RadioHound,
    )
    assert len(captures) > 0, "At least one capture should be present"
    for capture in captures:
        assert capture.uuid is not None, "Capture UUID should not be None"
        assert capture.capture_type == CaptureType.RadioHound
        assert not capture.channel, "RadioHound capture should not have a channel"
        assert capture.top_level_dir is not None, "Top level dir should not be None"


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.capture_listing(),
            *PassthruEndpoints.capture_creation(),
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
        ]
    ],
    indirect=True,
)
def test_capture_listing_all(integration_client: Client) -> None:
    """Tests reading and listing all types of captures."""
    # create a capture if none are present
    captures = integration_client.captures.listing()
    if not captures:
        _create_rh_capture(
            rh_sample_top_level_dir=dir_integration_data / "captures" / "radiohound",
            integration_client=integration_client,
        )
        captures = integration_client.captures.listing()
    assert len(captures) > 0, "At least one capture should be present"
    for capture in captures:
        assert capture.uuid is not None, "Capture UUID should not be None"
        assert capture.capture_type in CaptureType.__members__.values(), (
            "Capture type should be one of the known types: "
            f"{CaptureType.__members__.values()}"
        )
        assert capture.top_level_dir is not None, "Top level dir should not be None"
        assert capture.created_at is not None, (
            "Capture created_at timestamp should not be None"
        )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
        ]
    ],
    indirect=True,
)
def test_capture_update_rh(
    integration_client: Client,
    rh_sample_top_level_dir: Path,
) -> None:
    """Tests updating a RadioHound capture."""

    # ARRANGE by uploading capture data and creating one
    radiohound_data = load_rh_data(rh_sample_top_level_dir)

    # suffix path_after_capture_data with a random name to avoid conflicts between runs
    rh_capture_update_sds_path = rh_sample_top_level_dir.relative_to(
        dir_integration_data
    )
    random_suffix = get_random_line(10, include_punctuation=False)
    rh_capture_update_sds_path = (
        PurePosixPath(rh_capture_update_sds_path) / f"test-{random_suffix}"
    )

    _upload_assets(
        integration_client=integration_client,
        sds_path=rh_capture_update_sds_path,
        local_path=rh_sample_top_level_dir,
    )

    # delete all captures with that scan group to avoid conflicts
    assert "scan_group" in radiohound_data, (
        "Expected 'scan_group' in the RadioHound data"
    )
    scan_group = radiohound_data["scan_group"]
    _delete_rh_captures_by_scan_group(
        integration_client=integration_client,
        scan_group=scan_group,
    )

    capture_top_level = PurePosixPath("/") / rh_capture_update_sds_path
    capture = integration_client.captures.create(
        top_level_dir=capture_top_level,
        scan_group=scan_group,
        capture_type=CaptureType.RadioHound,
    )

    # certify the capture was created
    assert capture.uuid is not None, "Capture UUID should not be None"
    assert capture.capture_type == CaptureType.RadioHound
    assert capture.top_level_dir == capture_top_level
    assert capture.capture_props, "Capture properties should not be empty"
    assert capture.capture_props == radiohound_data, (
        "Capture props doesn't match the reference data: \n"
        f"'{capture.capture_props}'\n!=\n'{radiohound_data}'"
    )

    # upload a new radiohound file
    new_dir_top_level = dir_integration_data / "captures" / "radiohound-update"
    new_radiohound_file = new_dir_top_level / "reference-v0-addendum.rh.json"
    assert new_radiohound_file.is_file(), (
        "New reference file should exist; check that "
        f"you have the right paths set: '{new_radiohound_file}'"
    )

    _upload_assets(
        integration_client=integration_client,
        sds_path=rh_capture_update_sds_path,
        local_path=new_dir_top_level,
    )

    # ACT by updating the capture

    # update the capture
    integration_client.captures.update(
        capture_uuid=capture.uuid,
    )

    # ASSERT capture was updated

    # reading the capture should now return 2 files
    read_capture = integration_client.captures.read(
        capture_uuid=capture.uuid,
    )
    assert read_capture.files, "Expected a list of files associated to this capture"
    num_files = 2  # the original file + the new one
    assert len(read_capture.files) == num_files, (
        "Expected 2 files associated to this capture after update"
    )


def load_rh_data(rh_top_level_dir: Path) -> dict[str, Any]:
    """Helper to load RadioHound data from a top-level directory."""
    radiohound_file = rh_top_level_dir / "reference-v0.rh.json"
    assert radiohound_file.is_file(), (
        "Reference file should exist; check that "
        f"you have the right paths set: '{radiohound_file}'"
    )
    with radiohound_file.open("r") as fp_json:
        return json.load(fp_json)


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
        ]
    ],
    indirect=True,
)
def test_capture_upload_drf(
    integration_client: Client,
    drf_sample_top_level_dir: Path,
) -> None:
    """Tests uploading and creating a Digital-RF capture in one operation."""
    # ARRANGE a new directory to upload as capture
    test_dir = drf_sample_top_level_dir
    random_suffix = get_random_line(10, include_punctuation=False)
    sds_path = f"/test-upload-capture-{random_suffix}"
    capture_type = CaptureType.DigitalRF

    # ACT by uploading the capture
    capture = integration_client.upload_capture(
        local_path=test_dir,
        sds_path=sds_path,
        capture_type=capture_type,
        channel=drf_channel,
    )

    # ASSERT capture was correctly created
    assert capture is not None
    assert capture.uuid is not None, "Capture UUID should not be None"
    assert capture.capture_type == capture_type
    assert capture.channel == drf_channel
    assert str(capture.top_level_dir) == sds_path
    assert capture.capture_props["is_continuous"] is True
    assert "start_bound" in capture.capture_props
    assert "custom_attrs" in capture.capture_props

    # Clean up
    integration_client.captures.delete(capture_uuid=capture.uuid)


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
        ]
    ],
    indirect=True,
)
def test_capture_upload_rh(integration_client: Client) -> None:
    """Tests uploading and creating a RadioHound capture in one operation."""
    # ARRANGE
    dir_top_level = dir_integration_data / "captures" / "radiohound"
    radiohound_file = dir_top_level / "reference-v0.rh.json"
    assert radiohound_file.is_file(), (
        f"Reference file should exist at '{radiohound_file}'"
    )

    with radiohound_file.open("r") as fp_json:
        radiohound_data = json.load(fp_json)

    scan_group = radiohound_data["scan_group"]
    _delete_rh_captures_by_scan_group(
        integration_client=integration_client,
        scan_group=scan_group,
    )
    random_suffix = get_random_line(10, include_punctuation=False)
    sds_path = f"/test-upload-rh-{random_suffix}"

    # ACT
    capture = integration_client.upload_capture(
        local_path=dir_top_level,
        sds_path=sds_path,
        capture_type=CaptureType.RadioHound,
        scan_group=scan_group,
    )

    # ASSERT
    assert capture is not None
    assert capture.uuid is not None, "Capture UUID should not be None"
    assert capture.capture_type == CaptureType.RadioHound
    assert str(capture.top_level_dir) == sds_path
    assert str(capture.scan_group) == scan_group

    # Test capture props match the reference data
    assert capture.capture_props == radiohound_data, (
        "Capture props doesn't match the reference data"
    )

    # Clean up
    integration_client.captures.delete(capture_uuid=capture.uuid)


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
        ]
    ],
    indirect=True,
)
def test_capture_upload_missing_required_fields_drf(
    integration_client: Client, drf_sample_top_level_dir: Path
) -> None:
    """When lacking required fields, a DRF capture creation will fail.

    Note the file uploads are still expected to succeed.
    """
    # ARRANGE to upload assets for a Digital-RF capture
    test_dir = drf_sample_top_level_dir
    random_suffix = get_random_line(10, include_punctuation=False)
    sds_path = PurePosixPath(f"/test-upload-capture-{random_suffix}")
    capture_type = CaptureType.DigitalRF
    _upload_assets(
        integration_client=integration_client,
        sds_path=sds_path,
        local_path=test_dir,
    )

    # ACT & ASSERT - Missing channel for DigitalRF
    with pytest.raises(CaptureError):
        integration_client.upload_capture(
            local_path=test_dir,
            sds_path=sds_path,
            capture_type=capture_type,
            # Missing required channel parameter
        )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
            *PassthruEndpoints.capture_reading(),
        ]
    ],
    indirect=True,
)
def test_capture_reading_drf(
    integration_client: Client,
    drf_sample_top_level_dir: Path,
) -> None:
    """Tests reading a specific Digital-RF capture."""

    # ARRANGE

    # define paths in the context of a capture
    dir_top_level = drf_sample_top_level_dir
    dir_channel = dir_top_level / drf_channel
    dir_metadata = dir_channel / "metadata"
    assert dir_metadata.is_dir(), (
        "Metadata directory should exist; check that "
        f"you have the right paths set: '{dir_metadata}'"
    )

    # suffix path_after_capture_data with a random name to avoid conflicts between runs
    path_after_capture_data = dir_top_level.relative_to(dir_integration_data)
    random_suffix = get_random_line(10, include_punctuation=False)
    path_after_capture_data = path_after_capture_data / f"test-{random_suffix}"

    _upload_assets(
        integration_client=integration_client,
        sds_path=path_after_capture_data,
        local_path=dir_top_level,
    )
    capture = integration_client.captures.create(
        top_level_dir=Path(f"/{path_after_capture_data}"),
        channel=drf_channel,
        capture_type=CaptureType.DigitalRF,
    )
    assert capture.uuid is not None, "UUID of new capture should not be None"

    # ACT
    read_capture = integration_client.captures.read(
        capture_uuid=capture.uuid,
    )

    # ASSERT

    # basic capture information
    assert read_capture.files, "Expected a list of files associated to this capture"
    assert read_capture.uuid == capture.uuid, "Capture UUID should match"
    assert read_capture.capture_type == CaptureType.DigitalRF
    assert read_capture.channel == drf_channel
    assert read_capture.top_level_dir == Path(f"/{path_after_capture_data}")

    # test capture properties
    assert read_capture.capture_props == capture.capture_props, (
        "Capture properties should match the created capture"
    )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
            *PassthruEndpoints.capture_deletion(),
        ]
    ],
    indirect=True,
)
def test_capture_deletion(
    integration_client: Client, drf_sample_top_level_dir: Path
) -> None:
    """Tests deleting a capture."""

    # ARRANGE
    cap_data = _upload_drf_capture_test_assets(
        integration_client=integration_client,
        drf_sample_top_level_dir=drf_sample_top_level_dir,
    )
    capture = integration_client.captures.create(
        top_level_dir=cap_data.capture_top_level,
        capture_type=CaptureType.DigitalRF,
        channel=cap_data.drf_channel,
    )

    # ACT
    assert capture.uuid is not None, "UUID of new capture should not be None"
    result = integration_client.captures.delete(capture_uuid=capture.uuid)

    # ASSERT
    assert result is True, "Capture deletion should return True"
    with pytest.raises(CaptureError):
        integration_client.captures.read(
            capture_uuid=capture.uuid,
        )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
            *PassthruEndpoints.capture_search(),
        ]
    ],
    indirect=True,
)
def test_capture_advanced_search_frequency_range(
    integration_client: Client,
    drf_sample_top_level_dir: Path,
) -> None:
    """Tests searching captures with a frequency range query."""
    # ARRANGE to create a capture
    cap_data = _upload_drf_capture_test_assets(
        integration_client=integration_client,
        drf_sample_top_level_dir=drf_sample_top_level_dir,
    )
    capture = integration_client.captures.create(
        top_level_dir=cap_data.capture_top_level,
        channel=cap_data.drf_channel,
        capture_type=CaptureType.DigitalRF,
    )
    center_freq = capture.capture_props["center_freq"]
    test_freq_lower = center_freq - 1_234_567
    test_freq_upper = center_freq + 1_000_000

    # ACT by searching the new capture
    field_path = "capture_props.center_freq"
    query_type = "range"
    filter_value = {"gte": test_freq_lower, "lte": test_freq_upper}

    matched_caps = integration_client.captures.advanced_search(
        field_path=field_path,
        query_type=query_type,
        filter_value=filter_value,
    )
    is_capture_found = any(match.uuid == capture.uuid for match in matched_caps)

    # ASSERT the capture was found
    assert capture.uuid is not None, "UUID of new capture should not be None"
    assert len(matched_caps) > 0, (
        "Should find at least one capture in the frequency range"
    )
    assert is_capture_found, (
        f"Should find our test capture (UUID: {capture.uuid}) in search results"
    )

    integration_client.captures.delete(capture_uuid=capture.uuid)


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_capture_test")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.capture_creation(),
            *PassthruEndpoints.capture_search(),
        ]
    ],
    indirect=True,
)
def test_capture_advanced_search_full_text_search(
    integration_client: Client, drf_sample_top_level_dir: Path
) -> None:
    """Tests searching captures with a full-text search query.

    Digital-RF capture props example for reference:

    ```
    'capture_props': {
        'samples_per_second': 2500000,
        'start_bound': 1719499740,
        'end_bound': 1719499741,
        'is_complex': True,
        'is_continuous': True,
        'center_freq': 1024000000,
        'bandwidth': 100000000,
        'custom_attrs': {
            'num_subchannels': 1,
            'index': 4298748970000000,
            'processing/channelizer_filter_taps': [],
            'processing/decimation': 1,
            'processing/interpolation': 1,
            'processing/resampling_filter_taps': [],
            'processing/scaling': 1.0,
            'receiver/center_freq': 1024000000.4842604,
            'receiver/clock_rate': 125000000.0,
            'receiver/clock_source': 'external',
            'receiver/dc_offset': False,
            'receiver/description': 'UHD USRP source using GNU Radio',
            'receiver/id': '172.16.20.43',
            'receiver/info/mboard_id': 'n310',
            'receiver/info/mboard_name': 'n/a',
            'receiver/info/mboard_serial': '31649FE',
            'receiver/info/rx_antenna': 'RX2',
            'receiver/info/rx_id': '336',
            'receiver/info/rx_serial': '3162222',
            'receiver/info/rx_subdev_name': 'Magnesium',
            'receiver/info/rx_subdev_spec': 'A:0 A:1',
            'receiver/iq_balance': '',
            'receiver/lo_export': '',
            'receiver/lo_offset': 624999.4039535522,
            'receiver/lo_source': '',
            'receiver/otw_format': 'sc16',
            'receiver/samp_rate': 2500000.0,
            'receiver/stream_args': '',
            'receiver/subdev': 'A:0',
            'receiver/time_source': 'external',
        },
    },
    ```

    """
    # ARRANGE
    cap_data = _upload_drf_capture_test_assets(
        integration_client=integration_client,
        drf_sample_top_level_dir=drf_sample_top_level_dir,
    )
    capture = integration_client.captures.create(
        top_level_dir=cap_data.capture_top_level,
        channel=cap_data.drf_channel,
        capture_type=CaptureType.DigitalRF,
    )

    # ACT
    field_path = "capture_props.custom_attrs.receiver/description"
    query_type = "match"  # 'match' for full-text search
    filter_value = "UHD USRP"

    matched_caps = integration_client.captures.advanced_search(
        field_path=field_path,
        query_type=query_type,
        filter_value=filter_value,
    )
    is_capture_found = any(match.uuid == capture.uuid for match in matched_caps)

    # ASSERT
    assert len(matched_caps) > 0, "Should find at least one capture with the tag"
    assert is_capture_found, (
        f"Could not find our test capture '{capture.uuid}' in search results"
    )

    if capture.uuid:
        integration_client.captures.delete(capture_uuid=capture.uuid)


def _upload_assets(
    integration_client: Client,
    sds_path: Path | PurePosixPath,
    local_path: Path,
) -> None:
    """Helper to upload a local directory to SDS and assert success."""
    log.debug(f"Uploading assets as '/{sds_path}'")
    upload_results = integration_client.upload(
        local_path=local_path,
        sds_path=sds_path,
        verbose=False,
    )
    success_results = [success for success in upload_results if success]
    failed_results = [success for success in upload_results if not success]
    assert len(failed_results) == 0, (
        f"No failed uploads should be present: {failed_results}"
    )
    log.debug(f"Uploaded {len(success_results)} assets.")


def _create_rh_capture(
    rh_sample_top_level_dir: Path,
    integration_client: Client,
) -> Capture:
    """Helper to create a RadioHound capture.

    Note this deletes all existing RH captures under this user with the same scan group.
    """
    # ARRANGE
    radiohound_data = load_rh_data(rh_sample_top_level_dir)
    scan_group = radiohound_data["scan_group"]

    # suffix path_after_capture_data with a random name to avoid conflicts between runs
    rel_path_capture = rh_sample_top_level_dir.relative_to(dir_integration_data)
    random_suffix = get_random_line(10, include_punctuation=False)
    rel_path_capture = PurePosixPath(rel_path_capture) / f"test-{random_suffix}"

    _upload_assets(
        integration_client=integration_client,
        sds_path=rel_path_capture,
        local_path=rh_sample_top_level_dir,
    )

    # delete all captures with that scan group to avoid conflicts
    _delete_rh_captures_by_scan_group(
        integration_client=integration_client,
        scan_group=scan_group,
    )

    capture_top_level = PurePosixPath("/") / rel_path_capture
    return integration_client.captures.create(
        top_level_dir=capture_top_level,
        scan_group=scan_group,
        capture_type=CaptureType.RadioHound,
    )


def _delete_rh_captures_by_scan_group(
    integration_client: Client, scan_group: str
) -> None:
    """Helper to delete all RadioHound captures with a specific scan group."""
    captures = integration_client.captures.listing(
        capture_type=CaptureType.RadioHound,
    )
    same_scan_group_caps = [
        capture for capture in captures if str(capture.scan_group) == str(scan_group)
    ]
    if not same_scan_group_caps:
        log.debug("No captures to delete")
        return
    log.warning(
        f"Deleting {len(same_scan_group_caps)} captures with scan group '{scan_group}'"
    )
    for capture in same_scan_group_caps:
        log.warning(f"Deleting capture: {capture.uuid}")
        assert capture.uuid is not None, "Capture UUID should not be None"
        integration_client.captures.delete(capture_uuid=capture.uuid)


def __clean_all_captures(integration_client: Client) -> None:
    """Helper to delete all captures of this user in the SDS."""
    captures = integration_client.captures.listing()
    log.error(len(captures))
    for cap in captures:
        if cap.uuid is None:
            continue
        integration_client.captures.delete(capture_uuid=cap.uuid)



"""Common test fixtures and utilities for the SDS SDK."""

import os
import random
import sys
import uuid
from collections.abc import Generator
from pathlib import Path
from typing import Any

import pytest
from loguru import logger as log
from spectrumx import enable_logging
from spectrumx.client import Client
from spectrumx.gateway import API_TARGET_VERSION
from spectrumx.models.files import File
from spectrumx.ops import files
from spectrumx.utils import get_random_line

# better traceback formatting with rich, if available
try:
    from rich import traceback

    traceback.install()
except ImportError:
    log.warning("Install rich for better tracebacks")

enable_logging()

# Platform Specific Testing
PLATFORMS = {"darwin", "linux", "win32"}


def pytest_runtest_setup(item) -> None:
    supported_platforms = PLATFORMS.intersection(
        mark.name for mark in item.iter_markers()
    )
    plat = sys.platform
    if supported_platforms and plat not in supported_platforms:
        pytest.skip(f"cannot run on platform {plat}")


# ==== fixtures


@pytest.fixture
def client() -> Client:
    """Fixture to create a Client instance for testing."""
    return Client(host="sds-dev.crc.nd.edu")


@pytest.fixture
def sample_config() -> dict[str, Any]:
    """Fixture to create a sample configuration."""
    return {
        "SDS_SECRET_TOKEN": "12345",
        "HTTP_TIMEOUT": 42,
        "SDS_HOST": "sds-test.example.com",
    }


@pytest.fixture
def temp_file_empty(tmp_path: Path) -> Path:
    """Fixture to create a temporary file for testing."""
    file_path = tmp_path / "empty_test_file.txt"
    file_path.touch()
    return file_path


@pytest.fixture
def temp_file_with_text_contents(tmp_path: Path) -> Generator[Path]:
    """Fixture for a file with contents."""
    file_path = (
        tmp_path / f"test_file_{get_random_line(6, include_punctuation=False)}.txt"
    )
    # generate random string to change checksums every time
    with file_path.open("w", encoding="utf-8") as file_handle:
        file_handle.writelines(file_content_generator())

    yield file_path

    file_path.unlink(missing_ok=True)


@pytest.fixture
@pytest.mark.slow
def temp_large_binary_file(
    request: pytest.FixtureRequest,
    tmp_path: Path,
) -> Generator[Path]:
    """Fixture to create a temporary large binary file."""
    target_size_mb = (
        request.param.get("size_mb", None) if hasattr(request, "param") else None
    )
    if target_size_mb is None:
        target_size_mb = random.randint(10, 20)  # noqa: S311
    log.warning(f"Creating a large binary file of {target_size_mb:,} MB.")
    large_binary_file = tmp_path / "large_binary_file"
    byte_generator = random_bytes_generator(1024 * 1024 * target_size_mb)
    with large_binary_file.open("wb") as f_ptr:
        for chunk in byte_generator:
            f_ptr.write(chunk)

    yield large_binary_file

    large_binary_file.unlink(missing_ok=True)


@pytest.fixture
def temp_file_tree(
    tmp_path: Path,
    num_dirs: int = 3,
    num_files_per_dir: int = 3,
) -> Generator[Path]:
    """Fixture to create a temporary directory with files in subdirs."""

    _total_num_files = num_dirs * num_files_per_dir
    _all_created_files: list[Path] = []
    _all_created_dirs: list[Path] = []
    extension: str = "txt"

    for dir_idx in range(num_dirs):
        subdir = tmp_path / f"subdir_{dir_idx}"
        subdir.mkdir(exist_ok=True)
        _all_created_dirs.append(subdir)
        for file_sub_idx in range(num_files_per_dir):
            file_path = subdir / f"test_file_{file_sub_idx}.{extension}"
            _all_created_files.append(file_path)
            with file_path.open("w", encoding="utf-8") as file_handle:
                file_handle.writelines(file_content_generator())
    assert len(_all_created_files) == _total_num_files

    yield tmp_path

    # cleanup
    for file_path in _all_created_files:
        file_path.unlink(missing_ok=True)
    for dir_path in _all_created_dirs:
        dir_path.rmdir()


class FakeFileFactory:
    """Say that three times fast."""

    def __init__(
        self, num_files: int, *, tmp_path: Path, create_local: bool = False
    ) -> None:
        self.num_files: int = num_files
        self.create_local: bool = create_local
        self.fake_files: list[File] = []
        self.tmp_path: Path = tmp_path
        self.value: list[File] = []

    def _inner_loop(self) -> Generator[File, None, list[File]]:
        """Inner loop to generate fake files."""
        for _ in range(self.num_files):
            file_obj = files.generate_sample_file(uuid.uuid4())
            if self.create_local:
                file_obj.local_path = self.tmp_path / f"{file_obj.name}"
                file_obj.local_path.touch()
            self.fake_files.append(file_obj)
            yield file_obj
        self.value = self.fake_files
        return self.value

    def __iter__(self) -> Generator[File, None, list[File]]:
        """Sets self.value properly when this generator is nested."""
        self.value = yield from self._inner_loop()
        return self.value


@pytest.fixture
# yes, this returns a generator of a generator
def fake_files(
    request: pytest.FixtureRequest, tmp_path: Path
) -> Generator[FakeFileFactory]:
    """Fixture to create a list of fake files."""

    _file_count = request.param.get("file_count", 4)
    _create_local = request.param.get("create_local", False)

    # yield to test
    file_gen = FakeFileFactory(
        num_files=_file_count, create_local=_create_local, tmp_path=tmp_path
    )
    yield file_gen

    # get returned list of files for cleanup
    fake_files: list[File] = file_gen.value
    assert isinstance(fake_files, list), (
        "Expected a list of File instances as generator value"
    )

    # cleanup
    log.error(f"Cleaning up {len(fake_files)} fake files.")
    for file_obj in fake_files:
        assert isinstance(file_obj, File), "Expected a File instance"
        if file_obj.local_path:
            file_obj.local_path.unlink(missing_ok=True)


# ==== helpers


def file_content_generator(
    num_lines: int = 100, chars_per_line: int = 80
) -> Generator[str]:
    """Generator for file content."""
    for _current_line in range(num_lines):
        random_line = get_random_line(chars_per_line)
        yield f"{random_line}\n"


def get_captures_endpoint(
    client: Client,
    capture_id: str | None = None,
) -> str:
    """Returns the endpoint for captures."""
    base_endpoint = client.base_url + f"/api/{API_TARGET_VERSION}/assets/captures/"
    if capture_id:
        return base_endpoint + f"{capture_id}/"
    return base_endpoint


def get_files_endpoint(client: Client) -> str:
    """Returns the endpoint for the files API, with trailing slash."""
    return client.base_url + f"/api/{API_TARGET_VERSION}/assets/files/"


def get_content_check_endpoint(client: Client) -> str:
    """Returns the endpoint for the content check API."""
    return (
        client.base_url
        + f"/api/{API_TARGET_VERSION}/assets/utils/check_contents_exist/"
    )


def random_bytes_generator(size: int, chunk: int = 1024) -> Generator[bytes]:
    """Generates random binary data for tests."""
    for _ in range(size // chunk):
        yield os.urandom(chunk)
    yield os.urandom(size % chunk)



"""Data models for the SpectrumX Data System SDK."""

from spectrumx.models.files.file import File
from spectrumx.models.files.file import FileUpload
from spectrumx.models.files.permission import PermissionRepresentation
from spectrumx.models.files.permission import UnixPermissionStr

__all__ = ["File", "FileUpload", "PermissionRepresentation", "UnixPermissionStr"]



"""Dataset model for SpectrumX."""

from pydantic import UUID4
from pydantic import BaseModel


class Dataset(BaseModel):
    """A dataset in SDS."""

    uuid: UUID4 | None = None


__all__ = [
    "Dataset",
]



"""File model for SpectrumX."""

from datetime import datetime
from multiprocessing import RLock
from multiprocessing.synchronize import RLock as RLockT
from pathlib import Path
from pathlib import PurePosixPath
from typing import Annotated

from pydantic import ConfigDict
from pydantic import Field

from spectrumx import utils
from spectrumx.models.base import SDSModel
from spectrumx.models.files.permission import PermissionRepresentation
from spectrumx.models.files.permission import UnixPermissionStr


class File(SDSModel):
    """A file in SDS.

    Attributes:
        # shared attributes with SDS
        created_at:         The timestamp when the file was created.
        directory:          The path to the file, relative to the user storage in SDS.
        expiration_date:    The date when the file will be marked for deletion from SDS.
        media_type:         The MIME type of the file.
        name:               The user-defined name for this file.
        permissions:        The permissions for the file.
        size:               The size of the file in bytes.
        sum_blake3:         The BLAKE3 checksum of the file.
        updated_at:         The timestamp when the file was last updated.
        uuid:               The unique identifier of the file in SDS.

        # local attributes:
        is_sample:          Sample files are not written to disk (used in dry-run mode)
        local_path:         The path to the file on the local filesystem (includes name)
    """

    created_at: datetime
    directory: Annotated[PurePosixPath, Field(default_factory=PurePosixPath)]
    expiration_date: datetime | None
    media_type: str
    name: str
    permissions: UnixPermissionStr
    size: int
    updated_at: datetime

    is_sample: bool = False
    local_path: Path | None = None

    # events and state
    _is_downloading: bool = False
    _is_uploading: bool = False
    contents_lock: RLockT = Field(default_factory=RLock, exclude=True)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def model_post_init(self, _) -> None:
        """Post-initialization steps."""

    @property
    def path(self) -> PurePosixPath:
        """Returns the path to the file, relative to the owner's root on SDS."""
        return self.directory / self.name

    @property
    def is_local(self) -> bool:
        """Checks if the file contents are available locally."""
        if self._is_downloading:
            return False
        return (
            self.local_path is not None
            and self.local_path.exists()
            and self.local_path.is_file()
        )

    def compute_sum_blake3(self) -> str | None:
        """Calculates the BLAKE3 checksum of the file.

        If the file is currently being downloaded, this method will
            block until the download is complete or terminated.

        Args:
            block: when True, waits until the file contents are unlocked.
        Returns:
            The BLAKE3 checksum of the file,
                OR None if the file is not available locally.
        """
        # we deliberately don't cache the checksum because the file
        # might be changed externally by a different process or the user.
        with self.contents_lock:
            # content downloads cannot start within this block
            if not self.is_local:
                return None
            # this should not happen anyway, so let's just assert
            assert self.local_path is not None, "Local path is not set"
            return utils.sum_blake3(self.local_path)

    @property
    def chmod_props(self) -> str:
        """Converts human-readable permissions to chmod properties."""
        return self.model_dump(
            include={"permissions": True},
            context={"mode": PermissionRepresentation.OCTAL},
        )["permissions"]

    def is_same_contents(self, other: "File", *, verbose: bool = False) -> bool:
        """Checks if the file has the same contents as another local file."""
        this_sum = self.compute_sum_blake3()
        other_sum = other.compute_sum_blake3()
        if verbose:
            utils.log_user(f"{this_sum}  {self.local_path}")
            utils.log_user(f"{other_sum}  {other.local_path}")
        return this_sum == other_sum


class FileUpload(SDSModel):
    name: str | None
    directory: str
    media_type: str
    permissions: UnixPermissionStr | None

    @staticmethod
    def from_file(file: File) -> "FileUpload":
        return FileUpload(
            name=file.name,
            directory=str(file.directory),
            media_type=file.media_type,
            permissions=file.permissions,
        )


__all__ = ["File", "FileUpload"]



"""Tests for the utils module."""

from pathlib import Path

import pytest
from spectrumx import utils


def test_permission_string_invalid_length() -> None:
    """Test that an invalid permission string raises an error."""
    invalid_len = [
        "rwxrwxrwxr",
        "rw-rwxrwxrwx",
        "rwxr-xr",
    ]
    for perm in invalid_len:
        with pytest.raises(AssertionError):
            utils.validate_file_permission_string(perm)


def test_permission_string_invalid_chars() -> None:
    """Test that an invalid permission string raises an error."""
    invalid_chars = [
        "rwxrwxrws",
        "rw_______",
        "rwxrwxrw?",
    ]
    for perm in invalid_chars:
        with pytest.raises(AssertionError):
            utils.validate_file_permission_string(perm)


def test_permission_string_out_of_order() -> None:
    """Test that an invalid permission string raises an error."""
    invalid_order = [
        "rwxrwxwxr",
        "wrxrwxrwx",
        "rwxrxwrwx",
        "--------r",
    ]
    for perm in invalid_order:
        with pytest.raises(AssertionError):
            utils.validate_file_permission_string(perm)


def test_permission_string_valid() -> None:
    """Test that a valid permission string passes."""
    valid_cases = [
        "rwxrwxrwx",
        "rw-rw-r--",
        "r--r--r--",
        "---------",
        "rwx------",
    ]
    for valid in valid_cases:
        utils.validate_file_permission_string(valid)


def test_into_bool_truthy() -> None:
    """Makes sure truthy values are converted to True."""
    truthy_values = [
        True,
        1,
        "true",
        "True",
        "TrUe",
        "TRUE",
        "ON",
        "1",
        "t",
        "y",
        "yes",
        "on",
        "enabled",
    ]
    for truthy in truthy_values:
        assert utils.into_human_bool(truthy) is True, f"{truthy} failed"


@pytest.mark.linux
@pytest.mark.darwin
def test_clean_local_path() -> None:
    """Test that a local path is cleaned."""
    test_cases = [
        {
            # our usual case
            "input": Path("/files/user@domain.com/clean/path"),
            "expected": Path("clean/path"),
        },
        {
            # still clean it when not in a top-level
            "input": Path("files/user@domain.ai/clean/path"),
            "expected": Path("clean/path"),
        },
        {
            # still clean it when not "files"
            "input": Path("___/user@domain.ai/clean/path"),
            "expected": Path("clean/path"),
        },
        {
            # no changes
            "input": Path("example.com/change/path"),
            "expected": Path("example.com/change/path"),
        },
        {
            # no changes
            "input": Path("user@domain.dev/still/clean/path"),
            "expected": Path("user@domain.dev/still/clean/path"),
        },
        {
            # no changes
            "input": Path("yet/another/path"),
            "expected": Path("yet/another/path"),
        },
    ]
    for case in test_cases:
        entered = case["input"]
        expected = case["expected"]
        actual = utils.clean_local_path(entered)
        assert actual == case["expected"], (
            f"Failed: {entered} -> {actual} != {expected}"
        )


def test_into_bool_falsey() -> None:
    """Makes sure falsey values are converted to False."""
    falsey_values = [
        False,
        0,
        "false",
        "False",
        "FalSe",
        "FALSE",
        "OFF",
        "0",
        "f",
        "n",
        "no",
        "off",
        "disabled",
    ]
    for falsey in falsey_values:
        assert utils.into_human_bool(falsey) is False, f"{falsey} failed"



"""SpectrumX SDK package."""

# ruff: noqa: E402  # imports not at top-level

# -------------------
# better traceback formatting with rich, if available
try:
    from rich import traceback  # pyright: ignore[reportMissingImports]

    traceback.install()
except ImportError:
    pass

# -------------------
# package metadata

import importlib.metadata

LIB_NAME: str = "spectrumx"

# disables the loguru logger if it is not imported
# TODO: consider using the standard logging module and
#       wrapping it with loguru only when running tests
try:
    from loguru import logger as log

    log.disable(LIB_NAME)
except ImportError:
    pass

__version__ = importlib.metadata.version(LIB_NAME)

# -------------------
# experimental features


class Experiments:
    """Methods that enable experimental features."""


experiments = Experiments()

# -------------------
# package imports

from spectrumx import models
from spectrumx.client import Client

# -------------------
# package level functions


def main() -> None:  # pragma: no cover
    """Enables loguru logger when running the module."""
    enable_logging()


def enable_logging() -> None:
    """Enables loguru logger."""
    try:
        log.enable(LIB_NAME)  # pyright: ignore[reportPossiblyUnboundVariable]
        log.info(f"Enabled logging for '{LIB_NAME}'")  # pyright: ignore[reportPossiblyUnboundVariable]
    except NameError:
        import logging

        logger = logging.getLogger(__name__)

        logger.warning("Install Loguru to enable additional spectrumx logging.")


# -------------------
# package exports

__all__ = [
    "Client",
    "__version__",
    "enable_logging",
    "experiments",
    "models",
]

if __name__ == "__main__":  # pragma: no cover
    main()



"""File operations for SpectrumX SDK."""

import mimetypes
import uuid
from collections.abc import Generator
from datetime import datetime
from datetime import timedelta
from pathlib import Path
from pathlib import PurePosixPath

from loguru import logger as log

from spectrumx.models.files import File
from spectrumx.utils import log_user
from spectrumx.utils import log_user_warning

_tz = datetime.now().astimezone().tzinfo


def _load_undesired_globs(ignore_file: Path | None = None) -> list[str]:
    """Returns a list of undesired glob patterns loaded from a .sds-ignore file.

    The .sds-ignore file is similar to a .gitignore, with one pattern per line.
    Comments are marked with a # and ignored on read.
    Simple wildcards are supported, but recursive ones (**) are not.
    Exclusions with a ! prefix are not supported.
    """
    undesired_basenames = []
    if ignore_file is None:
        ignore_file = Path(__file__).parent / ".sds-ignore"
    comment_indicator = "#"
    if ignore_file.is_file():
        with ignore_file.open("r") as f:
            undesired_basenames = [
                line.split(comment_indicator, maxsplit=1)[0].strip()
                for line in f.read().splitlines()
                if not line.lstrip().startswith(comment_indicator)
            ]
            undesired_basenames = sorted([line for line in undesired_basenames if line])
    else:
        log.info("No .sds-ignore file found")
    return undesired_basenames


DISALLOWED_GLOBS = _load_undesired_globs()


def get_file_media_type(file_path: Path) -> str:
    """Returns the media type of a file.

    Args:
        file_path: The path to the file.

    https://www.iana.org/assignments/media-types/media-types.xhtml#application
    https://developer.mozilla.org/en-US/docs/Web/HTTP/MIME_types
    """
    mime_type, _encoding = mimetypes.guess_type(file_path)
    return mime_type or "application/octet-stream"


def get_file_permissions(file_path: Path) -> str:
    """Returns the permissions of a file."""
    permissions = file_path.stat().st_mode
    binary_rep = f"{permissions:09b}"[-9:]
    permission_map = ["r", "w", "x"]
    return "".join(
        permission_map[i % 3] if binary_rep[i] == "1" else "-" for i in range(9)
    )


def get_file_created_at(file_path: Path) -> datetime:
    """Returns the creation timestamp of a file."""
    return datetime.fromtimestamp(file_path.stat().st_ctime, tz=_tz)


def get_file_updated_at(file_path: Path) -> datetime:
    """Returns the last update timestamp of a file."""
    return datetime.fromtimestamp(file_path.stat().st_mtime, tz=_tz)


def construct_file(file_path: Path, sds_path: Path | PurePosixPath) -> File:
    """Constructs a file instance from a local file. File has to exist on disk."""
    file_path = Path(file_path)
    if not file_path.exists():
        msg = f"File not found: {file_path}"
        raise FileNotFoundError(msg)
    return File(
        name=file_path.name,
        expiration_date=None,
        local_path=file_path,
        size=file_path.stat().st_size,
        directory=PurePosixPath(sds_path),
        media_type=get_file_media_type(file_path),
        created_at=get_file_created_at(file_path),
        updated_at=get_file_updated_at(file_path),
        permissions=get_file_permissions(file_path),
    )


def is_valid_file(
    file_path: Path, *, check_sds_ignore: bool = True
) -> tuple[bool, list[str]]:
    """Returns True if the path is a valid file.
    A similar check is also performed at the server side.

    About check_sds_ignore's glob matches, we use Pathlib.match();
        see their docs to know more about what's supported:
        https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.match

    Args:
        file_path:          The path to the file.
        check_sds_ignore:   Whether to check for undesired glob patterns.
    Returns:
        True/False whether the file is valid.
        Reasons for invalidity otherwise.
    """
    file_mime = get_file_media_type(file_path)
    disallowed_mimes = [
        "application/x-msdownload",  # .exe
        "application/x-msdos-program",  # .com
        "application/x-msi",  # .msi
    ]
    is_valid_mime = file_mime not in disallowed_mimes
    reasons: list[str] = []
    if check_sds_ignore and any(file_path.match(glob) for glob in DISALLOWED_GLOBS):
        reasons.append(
            f"Path matched one or more undesired glob patterns for '{file_path.name}'"
        )
    if not is_valid_mime:
        reasons.append(f"Invalid MIME type: {file_mime}")
    if not file_path.is_file():
        reasons.append("Not a file")
    if file_path.stat().st_size == 0:
        reasons.append("Empty file")
    final_decision = (
        file_path.is_file() and file_path.stat().st_size > 0 and is_valid_mime
    )
    return final_decision, reasons


def get_valid_files(local_path: Path, *, warn_skipped: bool = False) -> Generator[File]:
    """Yields valid SDS files in the given directory.

    Args:
        local_path: The path to the directory.
    Yields:
        File instances.
    """
    recursive_file_list = local_path.rglob("*")
    successful_files: int = 0
    ignored_files: int = 0
    for file_path in recursive_file_list:
        if not file_path.is_file():
            continue
        is_valid, reasons = is_valid_file(file_path)
        if not is_valid:
            ignored_files += 1
            if warn_skipped:
                log_user_warning(f"Skipping {file_path}:")
                for reason in reasons:
                    log_user_warning(f"\t- {reason}")
            continue
        try:
            successful_files += 1
            local_rel_path = file_path.relative_to(local_path).parent
            yield construct_file(
                file_path=file_path, sds_path=PurePosixPath(local_rel_path)
            )
        except FileNotFoundError:
            continue
    log_user(
        f"Discovered {successful_files:,}/{successful_files + ignored_files:,} "
        "valid files"
    )
    if warn_skipped:
        if ignored_files:
            log_user_warning(f"Ignored {ignored_files} invalid files in '{local_path}'")
        else:
            log_user(f"No invalid files found in '{local_path}'")


def generate_sample_file(uuid_to_set: uuid.UUID) -> File:
    """Generates a sample file for dry-run mode."""
    tz = datetime.now().astimezone().tzinfo
    created_at = datetime.now(tz=tz)
    updated_at = created_at
    expiration_date = datetime.now(tz=tz) + timedelta(days=30)
    trailing_uuid_hex = uuid_to_set.hex[-6:]

    sample_file = File(
        uuid=uuid_to_set,
        name=f"dry-run-{trailing_uuid_hex}.txt",
        media_type="text/plain",
        size=888,
        directory=PurePosixPath("./sds-files/dry-run/"),
        permissions="rw-rw-r--",
        created_at=created_at,
        updated_at=updated_at,
        expiration_date=expiration_date,
        is_sample=True,  # always True for sample files
    )
    assert (
        sample_file.is_sample is True  # pyright: ignore[reportPrivateUsage]
    ), "SDS internal error: generated file is not a sample"
    return sample_file


def generate_random_files(num_files: int) -> list[File]:
    """Calls generate_sample_file() to create a list of fabricated files."""
    sample_files = []
    for _ in range(num_files):
        file_obj = generate_sample_file(uuid.uuid4())
        sample_files.append(file_obj)
    return sample_files



"""API functions specific to files."""

# ruff: noqa: SLF001
# pyright: reportPrivateUsage=false

import os
import tempfile
import uuid
from enum import Enum
from enum import auto
from multiprocessing.synchronize import RLock
from pathlib import Path
from pathlib import PurePosixPath

from loguru import logger as log
from pydantic import UUID4

from spectrumx.client import Client
from spectrumx.errors import SDSError
from spectrumx.models.files import File
from spectrumx.ops import files
from spectrumx.ops.pagination import Paginator
from spectrumx.utils import log_user
from spectrumx.utils import log_user_warning

log.trace("Placeholder log to avoid reimporting or resolving unused import warnings.")


class FileUploadMode(Enum):
    """Modes for uploading files to SDS."""

    SKIP = auto()  # no upload or update needed
    UPDATE_METADATA_ONLY = auto()  # no file contents, update an existing file entry
    UPLOAD_CONTENTS_AND_METADATA = auto()  # create a new file uploading everything
    UPLOAD_METADATA_ONLY = auto()  # no file contents, create a new file entry
    # file contents are immutable, so there is no "UPDATE_CONTENTS_ONLY"


def get_file(client: Client, file_uuid: UUID4 | str) -> File:
    """Get a file instance by its ID. Only metadata is downloaded from SDS.

    Note this does not download the file contents from the server. File
        instances still need to have their contents downloaded to create
        a local copy - see `download_file()`.

    Args:
        file_uuid: The UUID of the file to retrieve.
    Returns:
        The file instance, or a sample file if in dry run mode.
    """

    uuid_to_set: UUID4 = (
        uuid.UUID(file_uuid) if isinstance(file_uuid, str) else file_uuid
    )

    if not client.dry_run:
        file_bytes = client._gateway.get_file_by_id(uuid=uuid_to_set.hex)
        return File.model_validate_json(file_bytes)

    log_user("Dry run enabled: a sample file is being returned instead")
    return files.generate_sample_file(uuid_to_set)


def download_file(
    *,
    client: Client,
    file_instance: File | None = None,
    file_uuid: UUID4 | str | None = None,
    to_local_path: Path | str | None = None,
    skip_contents: bool = False,
    warn_missing_path: bool = True,
) -> File:
    """Downloads a file from SDS: metadata and contents (unless skip_contents=True).

    Note this function always overwrites the local path of the file instance.
    """
    # prepare file instance
    if isinstance(file_instance, File):
        file_instance, valid_uuid, valid_local_path_or_none = (
            __extract_download_info_from_file_instance(
                file_instance=file_instance,
                to_local_path=to_local_path,
                warn_missing_path=warn_missing_path,
            )
        )
    # or fetch file info from SDS creating a new instance
    else:
        file_instance, valid_uuid, valid_local_path_or_none = (
            __pre_fetch_file_for_download(
                client=client,
                file_uuid=file_uuid,
                to_local_path=to_local_path,
                warn_missing_path=warn_missing_path,
            )
        )

    __download_file_contents_if_applicable(
        client=client,
        file_instance=file_instance,
        valid_uuid=valid_uuid,
        valid_local_path_or_none=valid_local_path_or_none,
        skip_contents=skip_contents,
    )
    return file_instance


def list_files(
    *,
    client: Client,
    sds_path: PurePosixPath | Path | str,
    verbose: bool = False,
) -> Paginator[File]:
    """Lists files in a given SDS path.

    Args:
        sds_path: The virtual directory on SDS to list files from.
    Returns:
        A paginator for the files in the given SDS path.
    """
    sds_path = PurePosixPath(sds_path)
    if client.dry_run:
        log_user("Dry run enabled: files will be simulated")
    pagination: Paginator[File] = Paginator(
        gateway=client._gateway,
        sds_path=sds_path,
        Entry=File,
        dry_run=client.dry_run,
        verbose=verbose,
    )

    return pagination


def upload_file(
    *,
    client: Client,
    local_file: File | Path | str,
    sds_path: PurePosixPath | Path | str = "/",
) -> File:
    """Uploads a file to SDS.

    If the file instance passed already has a directory set, `sds_path` will
    be prepended. E.g. given:
        `sds_path = 'baz'`; and
        `local_file.directory = 'foo/bar/'`, then:
    The file will be uploaded to `baz/foo/bar/` (under the user root in SDS) and
        the returned file instance will have its `directory` attribute updated
        to match this new path.

    Args:
        local_file:     The local file to upload.
        sds_path:       The virtual directory on SDS to upload the file to.
    Returns:
        The file instance with updated attributes, or a sample when in dry run.
    """
    # validate inputs
    if not isinstance(local_file, (File, Path, str)):
        msg = f"file_path must be a Path, str, or File instance, not {type(local_file)}"
        raise TypeError(msg)
    local_file = Path(local_file) if isinstance(local_file, str) else local_file
    sds_path = PurePosixPath(sds_path)

    # construct the file instance if needed
    if isinstance(local_file, File):
        file_instance = local_file.model_copy()
        if file_instance.directory:
            composed_sds_path = sds_path / file_instance.directory
            file_instance.directory = composed_sds_path
    else:
        file_instance = files.construct_file(local_file, sds_path=sds_path)
    del local_file

    return __upload_file_mux(client=client, file_instance=file_instance)


def delete_file(client: Client, file_uuid: UUID4 | str) -> bool:
    """Deletes a file from SDS by its UUID.

    Args:
        client: The client instance.
        file_uuid: The UUID of the file to delete.
    Returns:
        True if the file was deleted successfully,
        or if in dry run mode (simulating success).
    Raises:
        SDSError: If the file couldn't be deleted.
    """
    uuid_to_delete: UUID4 = (
        uuid.UUID(file_uuid) if isinstance(file_uuid, str) else file_uuid
    )

    if client.dry_run:
        log_user(f"Dry run enabled: would delete file with UUID {uuid_to_delete.hex}")
        return True

    return client._gateway.delete_file_by_id(uuid=uuid_to_delete.hex)


def __download_file_contents_if_applicable(
    client: Client,
    file_instance: File,
    *,
    valid_uuid: UUID4,
    valid_local_path_or_none: Path | None,
    skip_contents: bool = False,
) -> None:
    if skip_contents:
        msg = (
            "A file instance was provided and skip_contents "
            "is True: nothing to download."
        )
        log_user_warning(msg)
        return
    if client.dry_run:
        file_instance.local_path = valid_local_path_or_none
        log_user(
            "Dry run enabled: file contents would be "
            f"downloaded as {file_instance.local_path}"
        )
        return
    downloaded_path: Path = __download_file_contents(
        client=client,
        file_uuid=valid_uuid,
        target_path=valid_local_path_or_none,
        contents_lock=file_instance.contents_lock,  # pyright: ignore[reportPrivateUsage]
    )
    file_instance.local_path = downloaded_path


def __download_file_contents(
    *,
    client,
    file_uuid: UUID4 | str,
    contents_lock: RLock,
    target_path: Path | None = None,
) -> Path:
    """Downloads the contents of a file from SDS to a location on disk.

    If target_path is not provided, a temporary file is created.
    When provided, the parent of target_path will be created if it does not exist.

    Args:
        file_uuid:      The UUID of the file to download from SDS.
        target_path:    The local path to save the downloaded file to.
    Returns:
        The local path to the downloaded file.
    """
    if target_path is None:
        file_desc, file_name = tempfile.mkstemp()
        os.close(file_desc)
        target_path = Path(file_name)
    target_path = Path(target_path)
    uuid_to_set: UUID4 = (
        uuid.UUID(file_uuid) if isinstance(file_uuid, str) else file_uuid
    )
    if not client.dry_run:
        target_path.parent.mkdir(parents=True, exist_ok=True)
        with target_path.open(mode="wb") as file_ptr, contents_lock:
            for chunk in client._gateway.get_file_contents_by_id(uuid=uuid_to_set.hex):
                file_ptr.write(chunk)
    else:
        log_user(f"Dry run enabled: file would be saved as {target_path}")
    return target_path


def __extract_download_info_from_file_instance(
    file_instance: File,
    *,
    to_local_path: Path | str | None,
    warn_missing_path: bool,
) -> tuple[File, UUID4, Path | None]:
    """Extracts from a file instance the info needed to download file contents."""
    if file_instance.uuid is None:
        msg = "The file passed is a local reference and cannot be downloaded."
        raise ValueError(msg)
    if to_local_path:
        file_instance.local_path = Path(to_local_path)
    if file_instance.local_path is None and warn_missing_path:
        msg = (
            "The file instance passed is missing a local path to "
            "download to. A temporary file will be created on disk."
        )
        log_user_warning(msg)
    valid_uuid = file_instance.uuid
    valid_local_path_or_none = file_instance.local_path
    return file_instance, valid_uuid, valid_local_path_or_none


def __get_upload_mode_and_asset(
    *, client: Client, file_instance: File
) -> tuple[FileUploadMode, uuid.UUID | None]:
    """Determines how to upload a file into SDS.

    Args:
        file_instance:  The file instance to get the upload mode for.
    Returns:
        The mode to upload the file in. Modes are:
            FileUploadMode.SKIP
            FileUploadMode.UPLOAD_CONTENTS_AND_METADATA
            FileUploadMode.UPLOAD_METADATA_ONLY
        Asset ID: The asset ID of the file in SDS, if it exists.

    SKIP is used when the file already exists in SDS and no changes are needed.
    UPLOAD_CONTENTS_AND_METADATA is used when the file is new or its contents
        have changed; it will create a new file entry with new contents.
    UPLOAD_METADATA_ONLY is used when the file contents are already in SDS, and
        it will create a new file entry while reusing existing contents from a
        sibling file.
    ---
    UPDATE_METADATA_ONLY is NOT returned here, as it might cause unintended changes.
        For this reason, "update" methods should be explicitly called by UUID.
    ---
    A "File" in SDS has immutable contents, so there is no "UPDATE_CONTENTS_ONLY".
    Note upload_* methods / modes create new assets in SDS, while update_* don't.
    """
    # always assume upload is needed in dry-run, since we can't contact the server
    if client.dry_run:
        return (
            FileUploadMode.UPLOAD_CONTENTS_AND_METADATA,
            None,
        )

    file_contents_check = client._gateway.check_file_contents_exist(
        file_instance=file_instance
    )
    asset_id = file_contents_check.asset_id

    if file_contents_check.file_exists_in_tree:
        return FileUploadMode.SKIP, asset_id
    if file_contents_check.file_contents_exist_for_user:
        return FileUploadMode.UPLOAD_METADATA_ONLY, asset_id
    return FileUploadMode.UPLOAD_CONTENTS_AND_METADATA, asset_id


def __pre_fetch_file_for_download(
    client: Client,
    file_uuid: UUID4 | str | None,
    to_local_path: Path | str | None,
    *,
    warn_missing_path: bool,
) -> tuple[File, UUID4, Path | None]:
    """Pre-fetches file metadata from SDS to get info for downloading contents."""
    if file_uuid is None:
        msg = "Expected a file instance or UUID to download."
        raise ValueError(msg)
    if to_local_path is None and warn_missing_path:
        msg = "The file will be downloaded as temporary."
        log_user_warning(msg)

    valid_local_path_or_none: Path | None = (
        Path(to_local_path) if to_local_path else None
    )

    valid_uuid: UUID4 = (
        uuid.UUID(file_uuid) if isinstance(file_uuid, str) else file_uuid
    )
    if client.dry_run:
        log_user("Dry run enabled: a sample file is being returned instead")
        file_instance: File = files.generate_sample_file(valid_uuid)
    else:
        file_bytes: bytes = client._gateway.get_file_by_id(uuid=valid_uuid.hex)
        file_instance: File = File.model_validate_json(file_bytes)
    return file_instance, valid_uuid, valid_local_path_or_none


def __upload_file_mux(*, client: Client, file_instance: File) -> File:  # noqa: C901
    """Uploads a file instance to SDS, choosing the right upload mode."""
    file_path = file_instance.local_path
    # check whether sds already has this file for this user
    upload_mode, asset_id = __get_upload_mode_and_asset(
        client=client, file_instance=file_instance
    )
    verbose = client.verbose

    if client.dry_run:
        log_user(f"Dry run enabled: skipping upload of '{file_path}'")
        return file_instance

    match upload_mode:
        case FileUploadMode.SKIP:
            if verbose:
                log_user(f"Skipping upload of existing '{file_path}'")
            return file_instance
        case FileUploadMode.UPLOAD_CONTENTS_AND_METADATA:
            if verbose:
                log_user(f"Uploading contents and metadata for '{file_path}'")
            return __upload_contents_and_metadata(
                client=client, file_instance=file_instance
            )
        case FileUploadMode.UPLOAD_METADATA_ONLY:
            if verbose:
                log_user(f"Uploading only metadata for '{file_path}'")
                log.debug(
                    f"Uploading only metadata '{file_path}' with sibling '{asset_id}'"
                )
            if asset_id is None:
                msg = "Expected an asset ID when uploading metadata only"
                raise SDSError(msg)
            return __upload_new_file_metadata_only(
                client=client, file_instance=file_instance, sibling_uuid=asset_id
            )
        case FileUploadMode.UPDATE_METADATA_ONLY:  # pragma: no cover
            if verbose:
                log_user(f"Updating metadata for '{file_path}'")
                log.debug(f"Updating metadata '{file_path}' with asset '{asset_id}'")
            assert asset_id is not None, "Expected an asset ID when updating metadata"
            return __update_existing_file_metadata_only(
                client=client, file_instance=file_instance, asset_id=asset_id
            )
        case _:  # pragma: no cover
            msg = f"Unexpected upload mode: {upload_mode}"
            raise SDSError(msg)


def __upload_contents_and_metadata(
    *,
    client: Client,
    file_instance: File,
) -> File:
    """UPLOADS a new file instance to SDS with contents and metadata.

    Args:
        client:         The SDS client instance.
        file_instance:  The file instance to upload.
    Returns:
        The file instance with updated attributes.
    """
    if client.dry_run:
        log_user("Dry run enabled: would upload contents and metadata")
        return file_instance

    assert not client.dry_run, "Internal error: expected dry run to be disabled."
    file_response = client._gateway.upload_new_file(file_instance=file_instance)
    uploaded_file = File.model_validate_json(file_response)
    uploaded_file.local_path = file_instance.local_path
    return uploaded_file


def __update_existing_file_metadata_only(
    *, client: Client, file_instance: File, asset_id: UUID4 | None
) -> File:
    """UPDATES an existing file instance with new metadata.

    ---
        Note: favor use of the safer `__upload_new_file_metadata_only()`,
        which creates a new File entry reusing the contents of a sibling file.
    ---

    Useful when the files is already instantiated in SDS and the file
        contents did not change: only the metadata needs to be updated.

    Args:
        file_instance:  The file instance with new metadata to update.
    Returns:
        The file instance with updated attributes.
    """
    if asset_id is not None:
        file_instance.uuid = asset_id

    if client.dry_run:
        msg = (
            "Dry run enabled: would update metadata "
            f"only for file '{file_instance.uuid}'"
        )
        log_user(msg)
        return file_instance

    assert not client.dry_run, "Internal error: expected dry run to be disabled."
    file_response = client._gateway.update_existing_file_metadata(
        file_instance=file_instance
    )
    updated_file = File.model_validate_json(file_response)
    updated_file.local_path = file_instance.local_path
    return updated_file


def __upload_new_file_metadata_only(
    *, client: Client, file_instance: File, sibling_uuid: UUID4
) -> File:
    """UPLOADS a new file instance to SDS, skipping the file contents.

    Useful when there are identical files in different locations: only their
        metadata needs to be uploaded to SDS. There must be a sibling file
        with the right contents, under the same user, already in SDS for this
        method to succeed.

    The sibling entry is not modified and doesn't need to exist locally.

    Args:
        file_instance:  The file instance with new metadata.
        sibling_uuid:   The UUID of the file which contents are identical to.
    Returns:
        The file instance with updated attributes.
    """
    if client.dry_run:
        log_user("Dry run enabled: uploading metadata only")
        return file_instance

    assert not client.dry_run, "Internal error: expected dry run to be disabled."
    file_response = client._gateway.upload_new_file_metadata_only(
        file_instance=file_instance, sibling_uuid=sibling_uuid
    )
    uploaded_file = File.model_validate_json(file_response)
    uploaded_file.local_path = file_instance.local_path
    return uploaded_file



"""API functions specific to captures."""

from __future__ import annotations

import json
import random
import uuid
from datetime import datetime
from pathlib import PurePosixPath
from typing import TYPE_CHECKING
from typing import Any
from uuid import uuid4

import pydantic
from loguru import logger as log

from spectrumx.errors import CaptureError
from spectrumx.models.captures import Capture
from spectrumx.models.captures import CaptureOrigin
from spectrumx.models.captures import CaptureType
from spectrumx.utils import log_user_warning

if TYPE_CHECKING:
    from pathlib import Path

    from spectrumx.gateway import GatewayClient

# python 3.10 backport
try:
    from datetime import UTC
except ImportError:
    from datetime import timezone

    UTC = timezone.utc  # noqa: UP017

index_mapping = {
    CaptureType.DigitalRF: "captures-drf",
    CaptureType.RadioHound: "captures-rh",
}


class CaptureAPI:
    gateway: GatewayClient
    verbose: bool = False

    def __init__(
        self,
        *,
        gateway: GatewayClient,
        dry_run: bool = True,
        verbose: bool = False,
    ) -> None:
        """Initializes the CaptureAPI."""
        self.dry_run = dry_run
        self.gateway = gateway
        self.verbose = verbose

    def create(
        self,
        *,
        top_level_dir: Path | PurePosixPath,
        capture_type: CaptureType,
        index_name: str = "",
        channel: str | None = None,
        scan_group: str | None = None,
    ) -> Capture:
        """Creates a new RF capture in SDS.

        An SDS capture is a collection of RF files that follow a predetermined structure
        which makes them suitable for indexing and searching. Example capture types
        include Digital-RF and RadioHound. The exact file structure and what files
        get indexed depend on the capture type.

        ### Notes:

        + This method doesn't upload files, it is a lightweight method that only groups
        files already in SDS into a capture. To upload files, see `upload()` and
        `upload_file()` from your SDS client instance.

        + The `top_level_dir` is the path in SDS relative to your user directory
        where the files were uploaded to, not your local filesystem path.

        Args:
            top_level_dir:  Virtual directory in SDS where capture files are stored.
            capture_type:   One of `spectrumx.models.captures.CaptureType`.
            index_name:     The SDS index name. Leave empty to automatically select.
            channel:        (For Digital-RF) the DRF channel name to index.
            scan_group:     (For RadioHound) UUIDv4 that groups RH files.
        Returns:
            The created capture object.
        Raises:
            CaptureError: If the capture couldn't be created e.g.: if it already exists,
                or the user doesn't have permission to create it.
        """
        if index_name:
            log.warning(
                "The 'index_name' parameter is deprecated and "
                "will be removed in future versions."
            )
        index_name = index_mapping.get(capture_type, index_name)
        if not index_name:
            log.warning(f"Could not find an index for {capture_type=}")

        top_level_dir = PurePosixPath(top_level_dir)
        if self.verbose:
            log.debug(
                f"Creating capture with {top_level_dir=}, "
                f"{channel=}, {capture_type=}, {index_name=}, {scan_group=}"
            )

        if self.dry_run:
            log.debug("Dry run enabled: simulating the capture creation")
            return Capture(
                capture_props={},
                capture_type=capture_type,
                channel=channel,
                index_name=index_name,
                origin=CaptureOrigin.User,
                scan_group=uuid.UUID(scan_group) if scan_group else None,
                top_level_dir=top_level_dir,
                uuid=uuid4(),
                files=[],
                created_at=datetime.now(UTC),
            )
        capture_raw = self.gateway.create_capture(
            capture_type=capture_type,
            channel=channel,
            index_name=index_name,
            scan_group=scan_group,
            top_level_dir=top_level_dir,
        )
        capture = Capture.model_validate_json(capture_raw)
        if self.verbose:
            log.debug(f"Capture created with UUID {capture.uuid}")
        return capture

    def listing(self, *, capture_type: CaptureType | None = None) -> list[Capture]:
        """Lists all RF captures in SDS under the current user.

        Note a capture must be manually "`.create()`-d" before it can be listed, even
            if all the files are uploaded to SDS.

        Args:
            capture_type:   The type of capture to list. If empty, it lists everything.
        Returns:
            A list of the RF captures found owned by the requesting user.
        """
        if self.verbose:
            log.debug(f"Listing captures of type {capture_type}")
        if self.dry_run:
            log.debug("Dry run enabled: simulating capture listing")
            num_captures: int = 3
            rng = random.Random()  # noqa: S311
            return [
                _generate_capture(
                    capture_type=capture_type
                    if capture_type
                    else rng.choice(list(CaptureType))
                )
                for _ in range(num_captures)
            ]
        captures_raw = self.gateway.list_captures(capture_type=capture_type)
        captures_list_raw, has_more = _extract_page_from_payload(captures_raw)
        if has_more:
            log.warning("Not all capture results may be listed. ")
            # TODO: request more pages if needed
        captures: list[Capture] = []
        for captures_raw in captures_list_raw:
            try:
                capture = Capture.model_validate(captures_raw)
                captures.append(capture)
            except pydantic.ValidationError as err:
                log_user_warning(f"Validation error loading capture: {captures_raw}")
                log.exception(err)
                continue
        if self.verbose:
            log.debug(f"Listing {len(captures)} captures")
        return captures

    def update(
        self,
        capture_uuid: uuid.UUID,
    ) -> None:
        """Updates a capture in SDS by re-discovering and re-indexing the files.

        Note that, just like the `.create()`, this method doesn't upload files.

        Args:
            capture_uuid:   The UUID of the capture to update.
        Returns:
            None
        """
        if self.verbose:
            log.debug(f"Updating capture with UUID {capture_uuid}")

        if self.dry_run:
            log.debug(f"Dry run enabled: simulating capture update {capture_uuid}")
            return

        capture_raw = self.gateway.update_capture(
            capture_uuid=capture_uuid,
            verbose=True,
        )
        capture = Capture.model_validate_json(capture_raw)
        if self.verbose:
            log.debug(f"Capture updated with UUID {capture.uuid}")

    def read(
        self,
        capture_uuid: uuid.UUID,
    ) -> Capture:
        """Reads a specific capture from SDS.

        This differs from the `.listing()` method by retrieving more information about
            the capture, including the associated files' UUID, directory, and name;
            useful when working with a specific capture.

        Args:
            capture_uuid:   The UUID of the capture to read.
        Returns:
            The capture object.
        """
        if self.verbose:
            log.debug(f"Reading capture with UUID {capture_uuid}")

        capture_raw = self.gateway.read_capture(capture_uuid=capture_uuid)
        capture = Capture.model_validate_json(capture_raw)
        if self.verbose:
            log.debug(f"Capture read with UUID {capture.uuid}")
        return capture

    def delete(self, capture_uuid: uuid.UUID) -> bool:
        """Deletes a capture from SDS by its UUID.

        Args:
            capture_uuid:   The UUID of the capture to delete.
        Returns:
            True if the capture was deleted successfully, or if in dry run mode.
        Raises:
            CaptureError: If the capture couldn't be deleted e.g.: if it doesn't exist;
                if it has already been deleted; if this user doesn't own it.
        """
        if self.verbose:
            log.debug(f"Deleting capture with UUID {capture_uuid}")

        if self.dry_run:
            log.debug(f"Dry run enabled: would delete capture {capture_uuid}")
            return True
        self.gateway.delete_capture(capture_uuid=capture_uuid)
        if self.verbose:
            log.debug(f"Capture deleted with UUID {capture_uuid}")
        return True

    def advanced_search(
        self: CaptureAPI,
        *,
        field_path: str,
        query_type: str,
        filter_value: str | dict[str, Any],
    ) -> list[Capture]:
        """Advanced searches for RF captures in SDS.

        SDS querying examples:
        https://github.com/spectrumx/sds-code/tree/master/gateway#opensearch-query-tips

        OpenSearch query DSL documentation:
        https://docs.opensearch.org/docs/latest/query-dsl/

        Args:
            field_path:     The field to search captures by e.g.:
                            `capture_props.center_freq`, `created_at`
            query_type:     The type of query to perform e.g.:
                            `range`, `geo_bounding_box`, `term`, `match`, `wildcard`
            filter_value:   The value(s) to filter by e.g.:
                            ```
                            {
                                "gte": 1990000000,
                                "lte": 2010000000
                            }
                            ```
        Returns:
            A list of captures matching the query.
        """

        # TODO: adapt this function to return a Paginator[Capture] object

        if self.dry_run:
            log.debug("Dry run enabled: simulating search results")
            rng = random.Random()  # noqa: S311
            return [
                _generate_capture(capture_type=rng.choice(list(CaptureType)))
                for _ in range(5)
            ]

        search_results_raw = self.gateway.captures_advanced_search(
            field_path=field_path,
            query_type=query_type,
            filter_value=filter_value,
        )
        search_results_obj = json.loads(search_results_raw)
        captures: list[Capture] = []
        if "results" not in search_results_obj:
            msg = "Unexpected search result format."
            raise CaptureError(msg)
        search_results = search_results_obj["results"]
        for result_raw in search_results:
            try:
                capture = Capture.model_validate(result_raw)
            except pydantic.ValidationError as err:
                log_user_warning(
                    f"Validation error loading search result: {result_raw}"
                )
                log.exception(err)
                continue
            else:
                captures.append(capture)
        if self.verbose:
            log.debug(f"Search returned {len(captures)} captures")
        return captures


def _extract_page_from_payload(
    capture_result_raw: bytes,
) -> tuple[list[dict[str, Any]], bool | None]:
    """Extracts the page from the payload.
    Args:
        capture_result_raw: The raw capture result from the API.
    Returns:
        The list of captures;
        A boolean indicating if there are more pages, or None if it can't be determined.
    """
    captures_object = json.loads(capture_result_raw)
    ret_captures_list: list[dict[str, Any]] = captures_object.get(
        "results", captures_object
    )

    # check if we need to request more pages
    has_more: bool | None
    if "next" not in ret_captures_list:
        has_more = None
    else:
        next_url: str = captures_object["next"]
        has_more = bool(next_url)

    # if result looks like a single capture, make sure it's a list
    if isinstance(ret_captures_list, dict):
        ret_captures_list = [ret_captures_list]

    return ret_captures_list, has_more


def _generate_capture(capture_type: CaptureType) -> Capture:
    """Generates a fake capture for testing purposes."""
    return Capture(
        capture_props={"_comment": "Simulated capture for a dry-run"},
        capture_type=capture_type,
        index_name="",
        channel=None,
        scan_group=None,
        origin=CaptureOrigin.User,
        top_level_dir=PurePosixPath("/"),
        uuid=uuid4(),
        files=[],
        created_at=datetime.now(UTC),
    )



"""Integration tests for file operations on SDS."""

import time
import uuid
from pathlib import Path
from pathlib import PurePosixPath
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest
from loguru import logger as log
from spectrumx.client import Client
from spectrumx.errors import FileError
from spectrumx.ops.files import construct_file
from spectrumx.ops.files import get_valid_files
from spectrumx.ops.files import is_valid_file
from spectrumx.utils import get_random_line

from tests.integration.conftest import PassthruEndpoints

if TYPE_CHECKING:
    from spectrumx.models.files import File

BLAKE3_HEX_LEN: int = 64


def test_is_valid_file_allowed(temp_file_with_text_contents) -> None:
    """Test the file validation function."""
    allowed_mime_types = [
        "application/json",
        "application/pdf",
        "application/x-hdf5",
        "application/xml",
        "image/gif",
        "image/jpeg",
        "image/png",
        "image/svg+xml",
        "text/csv",
        "text/plain",
    ]
    # mock get_file_media_type
    for mime_type in allowed_mime_types:
        with patch("spectrumx.ops.files.get_file_media_type", return_value=mime_type):
            is_valid, reasons = is_valid_file(temp_file_with_text_contents)
            assert is_valid, f"File with MIME type {mime_type} should be allowed."
            assert reasons == [], "No reasons should be given for valid files."


def test_is_valid_file_disallowed(temp_file_with_text_contents) -> None:
    """Test the file validation function."""
    disallowed_mime_types = [
        "application/x-msdownload",  # .exe
        "application/x-msdos-program",  # .com
        "application/x-msi",  # .msi
    ]
    # mock get_file_media_type
    for mime_type in disallowed_mime_types:
        with patch("spectrumx.ops.files.get_file_media_type", return_value=mime_type):
            is_valid, reasons = is_valid_file(temp_file_with_text_contents)
            assert not is_valid, (
                f"File with MIME type {mime_type} should be disallowed."
            )
            assert reasons, "Reasons should be given for disallowed files."


def test_get_valid_files(temp_file_tree: Path) -> None:
    """Test get_valid_files yields files matching glob."""

    all_file_paths = {path for path in temp_file_tree.rglob("*") if path.is_file()}
    assert all_file_paths, "No files found: can't run test."
    valid_file_instances = list(get_valid_files(temp_file_tree, warn_skipped=True))
    assert valid_file_instances, "No valid files found."

    # normalize paths to be relative to the temp_file_tree
    valid_file_paths = {
        Path(f"{temp_file_tree}/{file_instance.path}")
        for file_instance in valid_file_instances
    }

    # all test files should be valid and match the local path
    invalid_file_paths = all_file_paths - valid_file_paths
    assert invalid_file_paths == set(), (
        f"All files should be valid. Invalid paths: {invalid_file_paths}"
    )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
        ]
    ],
    indirect=True,
)
def test_upload_single_file(
    integration_client: Client, temp_file_with_text_contents: Path
) -> None:
    """Test file upload to SDS."""
    sds_path = PurePosixPath("/")
    local_file = construct_file(temp_file_with_text_contents, sds_path=sds_path)
    uploaded_file = integration_client.upload_file(
        local_file=temp_file_with_text_contents, sds_path=sds_path
    )
    assert uploaded_file.is_sample is False, "Sample file returned."
    for attr in uploaded_file.__dict__:
        log.debug(f"\t{attr:>15} = {uploaded_file.__dict__[attr]}")
    assert uploaded_file is not None, "File upload failed."
    assert isinstance(uploaded_file.uuid, uuid.UUID), "UUID not set."
    assert uploaded_file.size == local_file.size, "Size mismatch."
    assert len(uploaded_file.compute_sum_blake3() or "") == BLAKE3_HEX_LEN, (
        "Checksum not set."
    )
    assert uploaded_file.compute_sum_blake3() == local_file.compute_sum_blake3(), (
        "Checksum mismatch."
    )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
        ]
    ],
    indirect=True,
)
def test_upload_sibling(
    integration_client: Client, temp_file_with_text_contents: Path
) -> None:
    """Tests the sibling file upload (when file contents are already in the server)."""
    sds_path = Path("/")

    # create a copy of the temp file with text contents
    temp_file_copy = temp_file_with_text_contents.with_name(
        f"{temp_file_with_text_contents.stem}_COPY_{temp_file_with_text_contents.suffix}"
    )
    temp_file_copy.write_text(temp_file_with_text_contents.read_text())
    assert temp_file_copy.exists(), "Copy of the file should exist."
    assert temp_file_copy.is_file(), "Copy of the file should be a file."
    assert (
        temp_file_copy.stat().st_size == temp_file_with_text_contents.stat().st_size
    ), (
        "Copy of the file should have the same size as the original."
        f"{temp_file_copy.stat().st_size} != "
        "{temp_file_with_text_contents.stat().st_size}"
    )
    original_file = construct_file(temp_file_with_text_contents, sds_path=sds_path)
    copied_file = construct_file(temp_file_copy, sds_path=sds_path)

    # pre-conditions for test
    assert original_file.name != copied_file.name, "File names should be different."
    assert original_file.compute_sum_blake3() == copied_file.compute_sum_blake3(), (
        "Original and copied files should have the same checksum."
        f"{original_file.compute_sum_blake3()} != {copied_file.compute_sum_blake3()}"
    )

    # upload original file
    original_upload = integration_client.upload_file(
        local_file=temp_file_with_text_contents, sds_path=sds_path
    )

    # assertions of the original upload
    assert original_upload.local_path is not None, "Local path not set for original."
    assert original_upload.is_sample is False, "Sample file returned."
    for attr in original_upload.__dict__:
        log.debug(f"\t{attr:>15} = {original_upload.__dict__[attr]}")
    assert original_upload is not None, "File upload failed."
    assert isinstance(original_upload.uuid, uuid.UUID), "UUID not set."
    assert original_upload.size == original_file.size, "Size mismatch."
    assert original_upload.directory == original_file.directory, (
        "Directory should be the same."
        f"{original_upload.directory} != {original_file.directory}"
    )
    assert original_upload.name == original_file.name, (
        f"File names should be the same.{original_upload.name} != {original_file.name}"
    )

    # upload copy
    copy_upload = integration_client.upload_file(
        local_file=temp_file_copy, sds_path=sds_path
    )

    # assertions of the copy
    assert copy_upload.local_path is not None, "Local path not set for copy."
    assert copy_upload.is_sample is False, "Sample file returned."
    for attr in copy_upload.__dict__:
        log.debug(f"\t{attr:>15} = {copy_upload.__dict__[attr]}")
    assert copy_upload is not None, "File upload failed."
    assert isinstance(copy_upload.uuid, uuid.UUID), "UUID not set."
    assert copy_upload.directory == copied_file.directory, (
        "Directory should be the same."
        f"{copy_upload.directory} != {copied_file.directory}"
    )
    assert copy_upload.name == copied_file.name, (
        f"File names should be the same.{copy_upload.name} != {copied_file.name}"
    )

    # final assertions between files
    assert copy_upload.size == copied_file.size, "Size mismatch."
    assert copy_upload.compute_sum_blake3() == original_upload.compute_sum_blake3(), (
        "Checksum mismatch."
    )
    assert copy_upload.uuid != original_upload.uuid, (
        "UUIDs should NOT be the same for identical files."
        f"{copy_upload.uuid} != {original_upload.uuid}"
    )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
        ]
    ],
    indirect=True,
)
def test_upload_files_in_bulk(integration_client: Client, temp_file_tree: Path) -> None:
    """Tests uploading multiple files to SDS."""
    random_subdir_name = get_random_line(10, include_punctuation=False)
    results = integration_client.upload(
        local_path=temp_file_tree,
        sds_path=PurePosixPath("/test-tree") / random_subdir_name,
        verbose=True,
    )
    log.info(f"Uploaded {len(results)} files.")
    for upload_result in results:
        if upload_result:
            continue
        pytest.fail(f"File upload failed: {upload_result}")


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
        ]
    ],
    indirect=True,
)
@pytest.mark.parametrize(
    "temp_large_binary_file",
    [
        {"size_mb": 10},  # 10 MB
        {"size_mb": 100},  # 100 MB
        # {"size_mb": 1_000},  # 1 GB       # noqa: ERA001
        # {"size_mb": 10_000},  # 10 GB     # noqa: ERA001
    ],
    indirect=True,
)
def test_upload_large_file(
    integration_client: Client,
    temp_large_binary_file: Path,
) -> None:
    """Tests uploading a large file to SDS."""
    sds_path = PurePosixPath("/")
    local_file = construct_file(temp_large_binary_file, sds_path=sds_path)
    uploaded_file = integration_client.upload_file(
        local_file=temp_large_binary_file, sds_path=sds_path
    )
    assert uploaded_file.is_sample is False, "Sample file returned."
    assert uploaded_file is not None, "File upload failed."
    assert isinstance(uploaded_file.uuid, uuid.UUID), "UUID not set."
    assert uploaded_file.size == local_file.size, "Size mismatch."
    assert len(uploaded_file.compute_sum_blake3() or "") == BLAKE3_HEX_LEN, (
        "Checksum not set."
    )
    assert uploaded_file.compute_sum_blake3() == local_file.compute_sum_blake3(), (
        "Checksum mismatch."
    )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        PassthruEndpoints.file_content_checks(),
    ],
    indirect=True,
)
@pytest.mark.parametrize(
    "temp_large_binary_file",
    [
        {"size_mb": 1},
    ],
    indirect=True,
)
def test_check_file_content_non_existing(
    integration_client: Client,
    temp_large_binary_file: Path,
) -> None:
    """The file content checker must indicate new files don't exist in SDS."""
    file_instance = construct_file(temp_large_binary_file, sds_path=PurePosixPath("./"))
    file_contents_check = integration_client._gateway.check_file_contents_exist(  # pyright: ignore[reportPrivateUsage]  # noqa: SLF001
        file_instance
    )
    assert file_contents_check.file_contents_exist_for_user is False, (
        "Test file shouldn't exist for user."
    )
    assert file_contents_check.file_exists_in_tree is False, (
        "Test file shouldn't exist in tree."
    )
    assert file_contents_check.user_mutable_attributes_differ is True, (
        "Attributes should always differ for non-existent files."
    )
    assert file_contents_check.asset_id is None, (
        "Asset ID should be None for non-existent files."
    )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
        ]
    ],
    indirect=True,
)
def test_check_file_content_identical(
    integration_client: Client,
    temp_file_with_text_contents: Path,
) -> None:
    """The file content checker must indicate when files are identical."""
    random_string = get_random_line(10, include_punctuation=False)
    sds_path = PurePosixPath("/") / random_string
    file_instance = construct_file(temp_file_with_text_contents, sds_path=sds_path)

    # upload the file to sds
    uploaded_file = integration_client.upload_file(
        local_file=temp_file_with_text_contents, sds_path=sds_path
    )
    assert uploaded_file.uuid is not None, "UUID not set."
    # sleep
    time.sleep(2)

    file_contents_check = integration_client._gateway.check_file_contents_exist(  # pyright: ignore[reportPrivateUsage]  # noqa: SLF001
        file_instance
    )
    assert file_contents_check.file_contents_exist_for_user is True, (
        "Test file should exist for user."
    )
    assert file_contents_check.file_exists_in_tree is True, (
        "Test file should exist in tree."
    )
    assert file_contents_check.user_mutable_attributes_differ is False, (
        "Attributes should be identical."
    )
    assert file_contents_check.asset_id == uploaded_file.uuid, (
        "Asset ID does not match uploaded file: "
        f"{file_contents_check.asset_id} != {uploaded_file.uuid!s}"
    )


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
        ]
    ],
    indirect=True,
)
def test_check_file_content_name_changed(
    integration_client: Client,
    temp_file_with_text_contents: Path,
) -> None:
    """The file content checker must indicate when file names have changed."""
    random_string = get_random_line(10, include_punctuation=False)
    sds_path = PurePosixPath("/") / random_string

    # upload the file to sds
    uploaded_file = integration_client.upload_file(
        local_file=temp_file_with_text_contents, sds_path=sds_path
    )
    assert uploaded_file.uuid is not None, "UUID not set."

    # rename the file
    new_file_name = get_random_line(10, include_punctuation=False)
    new_file_path = temp_file_with_text_contents.with_name(new_file_name)
    assert new_file_path != temp_file_with_text_contents, "File name should change."
    Path.rename(
        temp_file_with_text_contents,
        new_file_path,
    )
    file_instance_renamed = construct_file(new_file_path, sds_path=sds_path)

    file_contents_check = integration_client._gateway.check_file_contents_exist(  # pyright: ignore[reportPrivateUsage]  # noqa: SLF001
        file_instance_renamed
    )
    assert file_contents_check.file_exists_in_tree is False, (
        "Test file shouldn't be identical to the one in SDS anymore."
    )
    assert file_contents_check.file_contents_exist_for_user is True, (
        "Test file contents should exist for this user, under a different name."
    )
    assert file_contents_check.user_mutable_attributes_differ is True, (
        "Attributes are different (name)."
    )
    assert file_contents_check.asset_id == uploaded_file.uuid, (
        "Expected asset ID to be the closest match (sibling UUID) to the uploaded file:"
        f"{file_contents_check.asset_id} != {uploaded_file.uuid!s}"
    )


# TODO:
# def test_file_upload_mode_skip(
# def test_file_upload_mode_contents_and_metadata(
# def test_file_upload_mode_metadata_only(


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_meta_download_or_upload(),
            *PassthruEndpoints.file_content_download(),
        ]
    ],
    indirect=True,
)
def test_download_single_file(
    integration_client: Client, temp_file_with_text_contents: Path, tmp_path: Path
) -> None:
    """Test file download from SDS."""
    sds_path = PurePosixPath("/")

    # upload a test file
    uploaded_file = integration_client.upload_file(
        local_file=temp_file_with_text_contents,
        sds_path=sds_path,
    )
    if uploaded_file is None or uploaded_file.uuid is None:
        pytest.fail("File upload failed.")

    # download that file to a different location
    download_path = tmp_path / "downloaded_file"

    try:
        downloaded_file = integration_client.download_file(
            file_uuid=uploaded_file.uuid, to_local_path=download_path
        )
        downloaded_file.is_same_contents(uploaded_file)

    # cleanup
    finally:
        download_path.unlink(missing_ok=True)


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_meta_download_or_upload(),
            *PassthruEndpoints.file_content_download(),
        ]
    ],
    indirect=True,
)
def test_download_files_in_bulk(
    integration_client: Client, temp_file_tree: Path
) -> None:
    """Test downloading multiple files from SDS."""
    random_subdir_name = get_random_line(10, include_punctuation=False)
    sds_path = PurePosixPath("/test-tree") / random_subdir_name
    results = integration_client.upload(
        local_path=temp_file_tree,
        sds_path=sds_path,
        verbose=True,
    )
    log.info(f"Uploaded {len(results)} files.")
    failures = [result for result in results if not result]
    if failures:
        pytest.fail(f"One or more file uploads failed: {failures}")
    uploaded_files: list[File] = [result() for result in results if result]
    assert len(uploaded_files) > 0, "No files uploaded."

    # download files into a subdirectory
    download_dir = temp_file_tree / "_downloaded_files"
    download_dir.mkdir(parents=True, exist_ok=True)
    downloaded_files_result = integration_client.download(
        from_sds_path=sds_path,
        to_local_path=download_dir,
        verbose=True,
    )
    downloaded_files = [result() for result in downloaded_files_result if result]
    download_failures = [
        result.error_info for result in downloaded_files_result if not result
    ]
    assert len(download_failures) == 0, f"Download failures: {download_failures}"

    # sort downloaded_files and uploaded_files by uuid
    downloaded_files = sorted(downloaded_files, key=lambda f: f.uuid or "-")
    uploaded_files = sorted(uploaded_files, key=lambda f: f.uuid or "-")
    assert len(downloaded_files) == len(uploaded_files), (
        "Number of downloaded files does not match the number of uploaded files."
    )

    for uploaded_file, downloaded_file in zip(
        uploaded_files, downloaded_files, strict=True
    ):
        download_path = downloaded_file.local_path
        upload_path = uploaded_file.local_path

        # check paths are valid
        assert download_path is not None, "Downloaded file not found."
        assert upload_path is not None, "Uploaded file not found."

        download_path = download_path.resolve()
        upload_path = upload_path.resolve()

        assert download_path.is_file(), "Downloaded file not found."
        assert upload_path.is_file(), "Downloaded file not found."

        # they must be different
        assert upload_path != download_path, (
            "Uploaded and downloaded file path should be different."
        )

        # check contents are the same
        assert downloaded_file.is_same_contents(uploaded_file, verbose=True), (
            f"Contents mismatch for file {uploaded_file.path.name}"
        )

        # assert downloaded path is a child of the download_dir
        assert download_path.is_relative_to(download_dir), (
            "Downloaded file path should be a child of the download directory."
            f"{download_path} not in {download_dir}"
        )

    log.info(f"Downloaded {len(uploaded_files)} files.")


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_meta_download_or_upload(),
            *PassthruEndpoints.file_content_download(),
        ]
    ],
    indirect=True,
)
def test_file_listing(integration_client: Client, temp_file_tree: Path) -> None:
    """Tests listing files in a directory on SDS."""
    random_subdir_name = get_random_line(10, include_punctuation=False)
    sds_path = PurePosixPath("/test-tree") / random_subdir_name
    results = integration_client.upload(
        local_path=temp_file_tree,
        sds_path=sds_path,
        verbose=True,
    )
    log.info(f"Uploaded {len(results)} files.")
    failures = [result for result in results if not result]
    if failures:
        pytest.fail(f"One or more file uploads failed: {failures}")
    uploaded_files: list[File] = [result() for result in results if result]
    assert len(uploaded_files) > 0, "No files uploaded."

    # list files in the directory
    integration_client.verbose = True
    files_listed = integration_client.list_files(sds_path=sds_path)
    listed_uuids = {file.uuid for file in files_listed}
    uploaded_uuids = {file.uuid for file in uploaded_files}
    assert listed_uuids == uploaded_uuids, "UUIDs mismatch."


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_uploads(),
            *PassthruEndpoints.file_meta_download_or_upload(),
            *PassthruEndpoints.file_deletion(),
        ]
    ],
    indirect=True,
)
def test_delete_file_integration(
    integration_client: Client, temp_file_with_text_contents: Path
) -> None:
    """Test file deletion by uploading a file then deleting it."""

    # make sure dry run is disabled
    integration_client.dry_run = False
    assert integration_client.dry_run is False, (
        "Dry run should be disabled for integration tests"
    )

    # ARRANGE

    # create a temporary file for testing
    temp_file_path = temp_file_with_text_contents
    temp_file_path.write_text("This is a test file to be deleted", encoding="utf-8")

    # upload the file
    sds_path = Path(f"/test-delete-file-{get_random_line(8)}")
    log.info(f"Uploading test file to SDS at {sds_path}")
    uploaded_file = integration_client.upload_file(
        local_file=temp_file_path, sds_path=sds_path
    )

    # verify file UUID exists
    assert uploaded_file.uuid is not None, "Uploaded file should have a UUID"
    file_uuid = uploaded_file.uuid
    log.info(f"File uploaded with UUID: {file_uuid}")

    # verify file exists by retrieving metadata
    retrieved_file = integration_client.get_file(file_uuid=file_uuid)
    assert retrieved_file.uuid == file_uuid, (
        "Retrieved file UUID should match uploaded file UUID"
    )

    # ACT

    # delete the file
    log.info(f"Deleting file with UUID: {file_uuid}")
    delete_result = integration_client.delete_file(file_uuid=file_uuid)

    # ASSERT

    assert delete_result is True, "File deletion should return True"
    with pytest.raises(FileError):
        integration_client.get_file(file_uuid=file_uuid)
    log.info(f"Test file '{file_uuid}' deleted successfully.")


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        [
            *PassthruEndpoints.file_content_checks(),
            *PassthruEndpoints.file_meta_download_or_upload(),
            *PassthruEndpoints.file_deletion(),
        ]
    ],
    indirect=True,
)
def test_delete_file_error_integration(integration_client: Client) -> None:
    """Test file deletion error when trying to delete a non-existent file."""
    # ARRANGE
    # Make sure dry run is disabled for integration testing
    integration_client.dry_run = False
    assert integration_client.dry_run is False, (
        "Dry run mode should be disabled for integration tests"
    )

    # Generate a random UUID that should not exist in the system
    non_existent_uuid = uuid.uuid4()
    log.info(f"Attempting to delete non-existent file with UUID: {non_existent_uuid}")

    # ACT & ASSERT
    # Try to delete a file with a non-existent UUID, which should raise a FileError
    with pytest.raises(FileError) as exc_info:
        integration_client.delete_file(file_uuid=non_existent_uuid)

    # ASSERT
    error_message = str(exc_info.value)
    log.info(f"FileError caught as expected: '{error_message}'")
    assert "not found" in error_message.lower() or "no file" in error_message.lower(), (
        "Error message should indicate the file was not found"
    )



"""SDS Client configuration."""

import logging
import os
from collections.abc import Callable
from collections.abc import Mapping
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import dotenv
from loguru import logger as log

from spectrumx.errors import Unset

from .models import SDSModel
from .utils import into_human_bool
from .utils import log_user
from .utils import log_user_warning

SDSModelT = type[SDSModel]
AttrValueT = str | int | float | bool
logger = logging.getLogger(__name__)


@dataclass
class Attr:
    """Attribute for the SDS configuration."""

    attr_name: str
    attr_value: AttrValueT | None = None
    cast_fn: Callable[[str], AttrValueT] | None = None


# '_cfg_name_lookup' maps config names to attribute names (internal).
#   This allows decoupling env file names from attribute
#   names in the object; used for refactoring code without
#   breaking user configuration files. Use lower case.
_cfg_name_lookup = {
    "dry_run": Attr(attr_name="dry_run", cast_fn=into_human_bool),
    "http_timeout": Attr(attr_name="http_timeout", cast_fn=int),
    "sds_host": Attr(attr_name="sds_host"),
    "sds_secret_token": Attr(attr_name="api_key"),
}


@dataclass
class DeprecatedOption:
    """Deprecated option for the SDS configuration."""

    deprecated_name: str
    new_name: str | None = None
    deprecation_version: str | None = None
    removal_version: str | None = None
    reason: str | None = None
    _user_warning: str | None = None

    @property
    def user_warning(self) -> str:
        """Gets the user warning."""
        warning = ""
        warning += f"Option '{self.deprecated_name}' is deprecated and will be removed"
        warning += f" in v{self.removal_version}." if self.removal_version else "."
        warning += f" Use '{self.new_name}' instead." if self.new_name else ""
        warning += f" {self.reason}" if self.reason else ""
        return warning


class SDSConfig:
    """Configuration for the SpectrumX Data System."""

    sds_host: None | str = None
    api_key: str = ""
    dry_run: bool = True  # safer default
    timeout: int = 30

    _active_config: list[Attr]
    _env_file: Path | None = None

    def __init__(
        self,
        *,
        env_file: Path | None = None,
        env_config: Mapping[str, Any] | None = None,
        sds_host: None | str = None,
        verbose: bool = False,
    ) -> None:
        """Initialize the configuration.
        Args:
            env_file:   Path to the environment file to load the config from.
            env_config: Overrides for the environment file.
            sds_host:   The host to connect to (the config file has priority over this).
            verbose:    Show which config files are loaded and which attributes are set.
        """
        self.sds_host = sds_host
        self.init_config(env_file=env_file, env_config=env_config, verbose=verbose)

    def init_config(
        self,
        *,
        env_file: Path | None | type[Unset] = Unset,
        env_config: Mapping[str, Any] | None,
        verbose: bool,
    ) -> None:
        """Initializes the client config, with `env_config` taking precedence.

        Args:
            env_file:   Path to the environment file to load the config from.
            env_config: Overrides for the environment file.
            verbose:    Show which config files are loaded and which attributes are set.

        """
        base_dir = Path.cwd()
        if env_file is Unset:
            self._env_file = Path(".env")
        elif isinstance(env_file, (str, Path)):
            self._env_file = Path(env_file)
        else:
            self._env_file = None
        if self._env_file and not self._env_file.is_absolute():
            self._env_file = base_dir / self._env_file
        clean_config = self.__load_config(env_cli_config=env_config, verbose=verbose)
        self._set_config(clean_config)

    def show_config(self, log_fn: Callable[[str], None] = print) -> None:
        """Show the active configuration."""
        header = "SDS_Config: active configuration:"
        log.debug(header)  # for sdk developers
        log_fn(header)  # for users
        for attr in self._active_config:
            _log_redacted(
                key=attr.attr_name,
                value=str(attr.attr_value),
                log_fn=log_fn,
            )

    def _set_config(self, clean_config: list[Attr]) -> None:
        """Sets the instance attributes."""
        for attr in clean_config:
            setattr(
                self,
                attr.attr_name,
                attr.value,
            )
            _log_redacted(key=attr.attr_name, value=attr.value)

        # validate attributes / show user warnings
        if not self.api_key:
            log.error(
                "SDS_Config: API key not set. Check your environment"
                " file has an SDS_SECRET_TOKEN set."
            )

        self._active_config = clean_config

    def __load_config(
        self,
        *,
        env_cli_config: Mapping[str, Any] | None = None,
        verbose: bool = False,
    ) -> list[Attr]:
        """Load the configuration."""
        if self._env_file and not self._env_file.exists():
            msg = f"Environment file missing: {self._env_file}"
            log_user_warning(msg)

        if verbose:
            if self._env_file:
                msg = f"SDS_Config: found environment file: {self._env_file}"
                log_user(msg)
            else:
                log_user("SDS_Config: not using an env file")

        # get variables from running env
        env_vars = {}
        if secret := os.environ.get("SDS_SECRET_TOKEN"):
            env_vars["SDS_SECRET_TOKEN"] = secret
        env_vars = {k: v for k, v in env_vars.items() if v is not None}
        log.debug(f"SDS_Config: from local env: {list(env_vars.keys())}")

        # merge file, cli, and env vars configs
        if (
            self._env_file
            and not self._env_file.exists()
            and self._env_file.name != ".env"
        ):
            log_user_warning(f"Custom env file not found: {self._env_file}")
        env_file_config = dotenv.dotenv_values(self._env_file, verbose=verbose)
        env_cli_config = env_cli_config or {}
        env_config = {**env_file_config, **env_cli_config, **env_vars}

        # clean and set the configuration loaded
        cleaned_config: list[Attr] = _clean_config(
            name_lookup=_cfg_name_lookup, env_config=env_config
        )

        # `deprecated_names` allows gracefully phasing out settings in future SDK
        # releases. Names are case-insensitive, but listed as uppercase for warnings
        # After `removal_version` is released, its entry may be removed from this list.
        deprecated_opts: list[DeprecatedOption] = [
            DeprecatedOption(
                deprecated_name="TIMEOUT",
                new_name="HTTP_TIMEOUT",
                deprecation_version="0.1.2",
                removal_version="0.2.0",
            ),
        ]
        # warn about deprecated options when found
        for dep_opt in deprecated_opts:
            if dep_opt.deprecated_name.lower() in env_config:
                log_user_warning(dep_opt.user_warning)

        return cleaned_config


def _clean_config(
    name_lookup: dict[str, Attr],
    env_config: Mapping[str, Any],
) -> list[Attr]:
    """Cleans the configuration to match known attributes."""
    cleaned_config: list[Attr] = []
    for key, sensitive_value in env_config.items():
        normalized_key: str = key.lower().replace("-", "_").replace(" ", "_")

        # skip if no value
        if sensitive_value is None:
            log.warning(f"SDS_Config: {key} has no value")
            continue
        attr: Attr | None = name_lookup.get(normalized_key)

        # warn of invalid config
        if attr is None:
            msg = f"SDS_Config: {key} not recognized"
            log.warning(msg)
            logger.warning(msg)
            continue

        # set value
        attr.value = attr.cast_fn(sensitive_value) if attr.cast_fn else sensitive_value

        # set the attribute, casting if necessary
        cleaned_config.append(attr)

    return cleaned_config


def _log_redacted(
    key: str,
    value: str,
    log_fn: Callable[[str], None] = logger.info,
    depth: int = 1,
) -> None:
    """Logs but redacts value if the key hints to a sensitive content, as passwords."""
    std_length: int = 4
    lower_case_hints = {
        "key",
        "secret",
        "token",
        "pass",
    }
    safe_value = (
        "*" * std_length
        if any(hint in key.lower() for hint in lower_case_hints)
        else value
    )
    del value
    msg = f"\tSDS_Config: set {key}={safe_value}"
    log.opt(depth=depth).debug(msg)  # for sdk developers
    log_fn(msg)


__all__ = ["SDSConfig"]



"""Integration tests with SDS for testing authentication.

Use the integration_client fixture for these tests.
"""

import pytest
from spectrumx import Client
from spectrumx.errors import AuthError

from tests.integration.conftest import PassthruEndpoints


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        PassthruEndpoints.authentication(),
    ],
    # tell pytest to pass the parameters to the fixture, \
    #   instead of the test function directly:
    indirect=True,
)
def test_authentication_200_succeeds(
    integration_client: Client,
) -> None:
    """Given a successful auth response, the client must be authenticated."""

    # auth test
    try:
        integration_client.authenticate()
        assert integration_client.is_authenticated
    except AuthError as err:
        msg = (
            "Authentication failed: make sure the api key in `integration.env` "
            "matches the server's: you need to set it manually. Error:"
            f" {err}"
        )
        pytest.fail(msg)


@pytest.mark.integration
@pytest.mark.usefixtures("_integration_setup_teardown")
@pytest.mark.usefixtures("_without_responses")
@pytest.mark.parametrize(
    "_without_responses",
    argvalues=[
        PassthruEndpoints.authentication(),
    ],
    # tell pytest to pass the parameters to the fixture, \
    #   instead of the test function directly:
    indirect=True,
)
def test_authentication_env_config_succeeds(
    inline_auth_integration_client: Client,
) -> None:
    """Tests that authentication with the inline dictionary succeeds."""

    # auth test
    inline_auth_integration_client._config.show_config()  # noqa: SLF001 # pyright: ignore[reportPrivateUsage]
    try:
        inline_auth_integration_client.authenticate()
        assert inline_auth_integration_client.is_authenticated
    except AuthError as err:
        msg = (
            "Authentication failed: make sure the api key in `integration.env` "
            "matches the server's: you need to set it manually. Error:"
            f" {err}"
        )
        pytest.fail(msg)



"""High-level usage tests for the SpectrumX client.

Find asset-specific tests in `./ops/`.
"""
# pylint: disable=redefined-outer-name

import uuid as uuidlib
from pathlib import Path
from pathlib import PurePosixPath

import pytest
from loguru import logger as log
from responses import RequestsMock
from spectrumx import Client
from spectrumx.errors import AuthError
from spectrumx.gateway import API_TARGET_VERSION

from tests.conftest import get_files_endpoint

# --------
# FIXTURES
# --------


@pytest.fixture
def responses_dry_run(responses: RequestsMock) -> RequestsMock:
    """Fixture to mock responses for dry run mode."""
    responses.assert_all_requests_are_fired = False
    return responses


def get_auth_endpoint(client: Client) -> str:
    """Returns the endpoint for the auth API, with trailing slash."""
    return client.base_url + f"/api/{API_TARGET_VERSION}/auth/"


# ------------------------
# TESTS FOR AUTHENTICATION
# ------------------------
def test_authentication_200_succeeds(client: Client, responses: RequestsMock) -> None:
    """Given a successful auth response, the client must be authenticated."""
    responses.get(
        get_auth_endpoint(client),
        body="{}",
        status=200,
        content_type="application/json",
    )
    # list registered URLs
    log.error(responses.calls)
    # disable the dry run mode for this,
    # since we're testing the actual request
    client.dry_run = False
    client.authenticate()
    assert client.is_authenticated


def test_authentication_401_fails(client: Client, responses: RequestsMock) -> None:
    """Given a failed auth response, the client must raise AuthError."""
    responses.get(
        get_auth_endpoint(client),
        body="{}",
        status=401,
        content_type="application/json",
    )
    client.dry_run = False  # to test the actual request
    with pytest.raises(AuthError):
        client.authenticate()


# ----------------------
# TESTS FOR DRY-RUN MODE
# ----------------------
class DryModeAssertionError(AssertionError):
    """Raised in test when a request is made in dry run mode."""


def test_dry_run_setter(client: Client) -> None:
    """Makes sure setter works, preventing unintended changes."""
    client.dry_run = False
    assert client.dry_run is False, "Dry-run setter failed."
    client.dry_run = True
    assert client.dry_run is True, "Dry-run setter failed."


def test_dry_run_enabled_by_default(client: Client) -> None:
    """Dry-run mode must be enabled by default."""
    assert client.dry_run is True, "Dry-run must be enabled by default."


def test_dry_auth_does_not_request(
    client: Client, responses_dry_run: RequestsMock
) -> None:
    """When in dry mode, the client must not make any requests."""
    responses_dry_run.add_callback(
        responses_dry_run.GET,
        url=get_auth_endpoint(client),
        callback=lambda _: DryModeAssertionError(
            "No requests must be made in dry run mode"
        ),
    )
    client.dry_run = True
    assert client.dry_run is True, "Dry-run setter failed."
    client.authenticate()


def test_dry_file_upload_does_not_request(
    client: Client, responses_dry_run: RequestsMock, temp_file_with_text_contents: Path
) -> None:
    """When in dry run mode, the upload method must not make any requests."""
    responses_dry_run.add_callback(
        responses_dry_run.POST,
        url=get_files_endpoint(client),
        callback=lambda _: DryModeAssertionError(
            "No requests must be made in dry run mode"
        ),
    )
    client.dry_run = True
    assert client.dry_run is True, "Dry run must be enabled for this test."
    _file_sample = client.upload_file(
        local_file=temp_file_with_text_contents,
        sds_path=PurePosixPath("/my/upload/location"),
    )


def test_dry_file_get_does_not_request(
    client: Client, responses_dry_run: RequestsMock
) -> None:
    """When in dry run mode, the get_file method must not make any requests."""

    file_id = uuidlib.uuid4()

    responses_dry_run.add_callback(
        responses_dry_run.GET,
        url=get_files_endpoint(client) + f"{file_id.hex}/",
        callback=lambda _: DryModeAssertionError(
            "No requests must be made in dry run mode"
        ),
    )

    client.dry_run = True
    assert client.dry_run is True, "Dry run must be enabled for this test."
    _file_sample = client.get_file(file_uuid=file_id.hex)



"""Pagination for SDS constructs."""

import json
import sys
import time
import uuid
from pathlib import Path
from pathlib import PurePosixPath
from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import TypeVar

# python 3.10 backport
if sys.version_info < (3, 11):  # noqa: UP036
    from typing_extensions import Self  # noqa: UP035 # Required backport
else:
    from typing import Self

from loguru import logger as log

from spectrumx.errors import FileError
from spectrumx.errors import Unset
from spectrumx.gateway import GatewayClient
from spectrumx.models import SDSModel
from spectrumx.ops import files

if TYPE_CHECKING:
    from collections.abc import Generator

T = TypeVar("T", bound=SDSModel)


class Paginator(Generic[T]):
    """Manages the state for paginating through files in SDS.

    A Paginator instance may be iterated over to fetch and return parsed
    entries automatically. Note that network calls will happen in the background
    and fetching requests happen once per page. Iterating it also consumes the
    generator, so any yielded content should be stored if needed in the future.

    ## Usage example

    ```
    file_paginator = Paginator[File](
        Entry=File,
        gateway=gateway,
        sds_path="/path/to/files",
        dry_run=False,
        verbose=True,
    )
    print(f"Total files matched: {len(file_paginator)}")
    # len() will fetch the first page

    for my_file in file_paginator:
        print(f"Processing file: {my_file.name}")
        process_file(my_file)
        # new pages are fetched automatically

    for _my_file in file_paginator:
        msg = "This will not run, as the paginator was consumed."
        raise AssertionError(msg)
    ```
    """

    def __init__(
        self,
        *,
        Entry: type[SDSModel],  # noqa: N803
        gateway: GatewayClient,
        sds_path: PurePosixPath | Path | str,
        dry_run: bool = False,
        page_size: int = 30,
        start_page: int = 1,
        total_matches: int | None = None,
        verbose: bool = False,
    ) -> None:
        """Initializes the paginator with the required parameters.

        Args:
            Entry:          The SDSModel subclass to use when parsing the entries.
            gateway:        The gateway client to use for fetching pages.
            sds_path:       The SDS path to paginate through.
            dry_run:        If True, will generate synthetic pages instead of fetching.
            page_size:      The number of entries to fetch per page.
            start_page:     The page number to start fetching from.
            total_matches:  The total number of entries across all pages.
            verbose:        If True, will log more information about the pagination.
        """

        # TODO: generalize this to any SDSModel subclass (too coupled to File now)

        if page_size <= 0:  # pragma: no cover
            msg = "Page size must be a positive integer."
            raise ValueError(msg)
        if not isinstance(start_page, int) or start_page < 1:  # pragma: no cover
            msg = "Start page must be a positive integer."
            raise ValueError(msg)
        if (
            not isinstance(total_matches, int) and total_matches is not None
        ):  # pragma: no cover
            msg = "Total matches must be an integer."
            raise ValueError(msg)
        if not isinstance(sds_path, (PurePosixPath, Path, str)):  # pragma: no cover
            msg = "SDS path must be a PurePosixPath, Path, or str."
            raise TypeError(msg)
        if not isinstance(gateway, GatewayClient):  # pragma: no cover
            msg = "Gateway client must be provided."
            raise TypeError(msg)
        if not issubclass(Entry, SDSModel):  # pragma: no cover
            msg = "Entry must be a subclass of SDSModel."
            raise TypeError(msg)
        self.dry_run = dry_run
        self._Entry = Entry
        self._gateway = gateway
        self._next_page = start_page
        self._page_size = page_size
        self._sds_path = PurePosixPath(sds_path)
        self._total_matches = total_matches if total_matches else 1
        self._verbose: bool = verbose

        # internal state
        self._has_fetched = False
        self._is_fetching: bool = False
        self._current_page_data: dict[str, Any] | None = None
        self._current_page_entries: Generator[T] = iter(())
        self._next_element: T | Unset = Unset()
        self._yielded_count: int = 0

    def __iter__(self) -> Self:
        """Returns the iterator object."""
        return self

    def __next__(self) -> T:
        """Returns the next entry in the pagination."""
        while self._next_element is not Unset or self._has_next_page:
            try:
                self._next_element = next(self._current_page_entries)
            except StopIteration:
                try:
                    self._fetch_next_page()
                    self._debug("Fetched the next page.")
                except StopIteration as err:
                    self._debug("No more pages available.")
                    raise StopIteration(err) from err
            else:
                self._yielded_count += 1
                return self._next_element
        # execution should never reach here
        msg = "Internal paginator error."  # pragma: no cover
        raise RuntimeError(msg)  # pragma: no cover

    def __len__(self) -> int:
        """Returns the total number of entries."""
        if not self._has_fetched:
            self._debug("Fetching the first page of results.")
            self._fetch_next_page()
        return self._total_matches

    def __bool__(self) -> bool:
        """Returns True if there are entries to iterate over."""
        if not self._has_fetched:
            self._debug("Fetching the first page of results.")
            self._fetch_next_page()
        return self._total_matches > 0

    @property
    def _total_pages(self) -> int:
        """Calculates the total number of pages."""
        return (self._total_matches // self._page_size) + 1

    def _debug(self, message: str, depth: int = 1) -> None:
        """Logs a debug message if verbose mode is enabled."""
        if self._verbose:  # pragma: no cover
            log.opt(depth=depth).debug(message)

    @property
    def _has_next_page(self) -> bool:
        """Checks if there is a next page available."""
        has_not_fetched = not self._has_fetched
        has_pages_left = self._next_page <= self._total_pages
        return has_not_fetched or has_pages_left

    def _fetch_next_page(self) -> None:
        """Fetches the next page of results."""
        if self._is_fetching:  # pragma: no cover
            log.warning("Already fetching the next page.")
            return
        self._is_fetching = True
        # try-finally to unset self._is_fetching when done
        try:
            if not self._has_next_page:
                msg = "No more pages available."
                raise StopIteration(msg)
            if self.dry_run:
                self._debug("Dry run enabled: generating a synthetic page")
                self._ingest_fake_page()
            else:
                # try to fetch the next page
                try:
                    raw_page = self._gateway.list_files(
                        sds_path=self._sds_path,
                        page=self._next_page,
                        page_size=self._page_size,
                        verbose=self._verbose,
                    )
                    self._ingest_new_page(raw_page)
                except FileError as err:  # pragma: no cover
                    # log an unexpected FileError if it happens
                    if "invalid page" not in str(err).lower():
                        msg = "Unexpected error while fetching the next page:"
                        log.exception(msg)
                    msg = "No more pages available."
                    raise StopIteration(msg) from err
            self._next_page += 1
            self._has_fetched = True  # `len(self)` is now valid
        finally:
            self._is_fetching = False

    def _ingest_fake_page(self) -> None:
        """Loads a fake page into memory for dry-run mode."""
        self._total_matches = int(self._page_size * 2.5)  # targeting 3 pages
        _remaining_matches = self._total_matches - self._yielded_count
        _page_length = min(self._page_size, _remaining_matches)
        if self._yielded_count >= self._total_matches:  # pragma: no cover
            msg = "No more entries available."
            raise StopIteration(msg)
        # generate page
        if issubclass(self._Entry, files.File):
            # TODO: merge self._current_page_entries with the new
            #   entries (lazily) to allow pre-fetching a page, or
            #   fetching multiple pages at once.
            self._current_page_entries = (
                files.generate_sample_file(uuid.uuid4()) for _ in range(_page_length)
            )
        else:  # pragma: no cover
            msg = f"Dry-run mode not implemented for this entry type: {self._Entry}"
            raise NotImplementedError(msg)

    def _ingest_new_page(self, raw_page: bytes) -> None:
        """Loads a raw page into memory and updates pagination state.

        The total number of pages and entry count might change between page
            requests, so this will update these control variables accordingly.
        """
        try:
            self._current_page_data = json.loads(raw_page)
        except json.JSONDecodeError as err:  # pragma: no cover
            msg = "Failed to load page data: failed to decode the JSON response."
            raise TypeError(msg) from err
        if not isinstance(self._current_page_data, dict):  # pragma: no cover
            msg = "Failed to load page data: expected a dictionary from JSON."
            raise TypeError(msg)
        if "count" in self._current_page_data:
            self._total_matches = self._current_page_data["count"]
        self._current_page_entries = (
            (
                self._Entry(**entry_data)
                for entry_data in self._current_page_data["results"]
            )
            if "results" in self._current_page_data
            else iter(())
        )


def _process_file_fake(my_file: files.File) -> None:  # pragma: no cover
    """Sleeps a bit."""
    time.sleep(0.07)


def main() -> None:  # pragma: no cover
    """Usage example for paginator."""
    log.info("Running the main script.")
    file_paginator = Paginator[files.File](
        Entry=files.File,
        gateway=GatewayClient(
            host="localhost",
            api_key="does-not-matter-in-dry-run",
        ),
        sds_path="/path/to/files",
        page_size=10,
        dry_run=True,  # in dry-run this should always generate 2.5 pages
        verbose=True,
    )
    log.info(f"Total files matched: {len(file_paginator)}")
    processed_count: int = 0
    for my_file in file_paginator:
        log.info(f"Processing file: {my_file.name}")
        _process_file_fake(my_file)
        processed_count += 1
        # new pages are fetched automatically

    log.info(f"Processed {processed_count} / {len(file_paginator)} files.")

    log.info("Trying another loop:")
    for _my_file in file_paginator:
        msg = "This will not run, as the paginator was consumed."
        raise AssertionError(msg)
    log.info("No more files to process.")

    log.info("Paginator demo finished.")


if __name__ == "__main__":
    main()



import sys

# python 3.10 backport
if sys.version_info < (3, 11):  # noqa: UP036
    from backports.strenum import StrEnum  # noqa: UP035 # Required backport
else:
    from enum import StrEnum

from typing import Annotated
from typing import Any

from pydantic import BeforeValidator
from pydantic import Field
from pydantic import PlainSerializer
from pydantic import SerializationInfo
from pydantic import StringConstraints
from pydantic import TypeAdapter

PermissionOct = Annotated[
    int,
    Field(
        ge=0,
        le=511,
        description="Octal representation of Unix file permissions (e.g., 0x755).",
        examples=["0o755", "0o644", "0o777"],
    ),
]

PermissionStr = Annotated[
    str,
    Field(
        description="Unix file permission string (e.g. 'rwxr-xr--')",
        examples=["rwxr-xr-x", "rw-r--r--", "rwx------"],
    ),
    StringConstraints(
        strip_whitespace=True, min_length=9, max_length=9, pattern=r"^[rwx-]{9}$"
    ),
]
PermissionStrAdapter = TypeAdapter(PermissionStr)


# Helper functions
def octal_to_unix_perm_string(value: int) -> str:
    """
    Convert an octal Unix file permission to a string representation.

    :param value: An int representing the file permission in octal format (e.g., 0o755)
    :return: A 9-character string representing Unix permissions (e.g., 'rwxr-xr-x')
    :raises ValueError: If the input is not a valid octal Unix permission.
    """
    perm_oct = TypeAdapter(PermissionOct).validate_python(value)

    mapping = {
        0: "---",
        1: "--x",
        2: "-w-",
        3: "-wx",
        4: "r--",
        5: "r-x",
        6: "rw-",
        7: "rwx",
    }

    return "".join(mapping[(perm_oct >> (3 * i)) & 0o7] for i in reversed(range(3)))


def unix_perm_string_to_octal(value: str) -> int | ValueError:
    """
    Convert a Unix permission string (e.g., 'rwxr-xr-x') to an octal integer.

    :param value: A 9-character Unix permission string (e.g., 'rw-r--r--')
    :return: An integer representing the octal value of the permission (e.g., 0o755)
    :raises ValueError: If the input string is not a valid Unix permission string.
    """
    perm_str = PermissionStrAdapter.validate_python(value)

    one_hot = "".join(["0" if flag == "-" else "1" for flag in perm_str])
    return int(one_hot, base=2)


def unix_perm_from_any(value: Any) -> str | ValueError:
    if isinstance(value, int):
        return octal_to_unix_perm_string(value)
    if isinstance(value, str):
        return PermissionStrAdapter.validate_python(value)

    msg = f"Invalid Unix permission format: {value}"
    return ValueError(msg)


class PermissionRepresentation(StrEnum):
    STRING = "string"
    OCTAL = "octal"

    @classmethod
    def _missing_(cls, value):
        """Raises an exception when an unknown value is provided."""
        msg = f"Invalid PermissionRepresentation value: {value}"
        raise ValueError(msg)

    def convert(self, value: str) -> str:
        if self == PermissionRepresentation.OCTAL:
            return f"0o{unix_perm_string_to_octal(value):03o}"
        return value


# Serialization
def serialize_unix_permission(value: Any, info: SerializationInfo):
    """Serialize Unix permission based on context (`octal` or `string`)."""

    # If we have no context return the string representation
    if info.context is None:
        return value

    # Get the mode from the context and default to string
    mode = info.context.get("mode", PermissionRepresentation.STRING)
    return PermissionRepresentation(mode).convert(value)


UnixPermissionStr = Annotated[
    PermissionOct | PermissionStr,
    BeforeValidator(unix_perm_from_any),
    PlainSerializer(serialize_unix_permission, return_type=str),
]

__all__ = ["PermissionRepresentation", "UnixPermissionStr"]
